from __future__ import annotations

import warnings
from enum import IntEnum
from typing import List

from boa3.neo3 import contracts
from boa3.neo3.core import types, IJson


class ContractParameterType(IntEnum):
    SIGNATURE = 0x00,
    BOOLEAN = 0x01,
    INTEGER = 0x02,
    HASH160 = 0x03,
    HASH256 = 0x04,
    BYTEARRAY = 0x05,
    PUBLICKEY = 0x06,
    STRING = 0x07,
    ARRAY = 0x10,
    MAP = 0x12,
    INTEROP_INTERFACE = 0xf0,
    ANY = 0xfe,
    VOID = 0xff


class ContractParameterDefinition(IJson):
    """
    A parameter description for a contract Method or Event.
    """

    def __init__(self, name: str, type: contracts.ContractParameterType):
        """
        Args:
            name: the human readable identifier.
            type: the type of parameter.
        """
        self.name = name
        self.type = type

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.name == other.name and self.type == other.type

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        json = {
            "name": self.name,
            "type": self.type.name.title()
        }
        return json

    @classmethod
    def from_json(cls, json: dict) -> ContractParameterDefinition:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary keys.
        """
        return cls(
            name=json['name'],
            type=contracts.ContractParameterType[json['type'].upper()]
        )


class ContractEventDescriptor(IJson):
    """
    A description for an event that a contract can broadcast.
    """

    def __init__(self, name: str, parameters: List[ContractParameterDefinition]):
        """

        Args:
            name: the human readable identifier of the event.
            parameters: the list of parameters the event takes.
        """
        self.name = name
        self.parameters = parameters

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.name == other.name and self.parameters == other.parameters

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        json = {
            "name": self.name,
            "parameters": list(map(lambda p: p.to_json(), self.parameters))
        }
        return json

    @classmethod
    def from_json(cls, json: dict) -> ContractEventDescriptor:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary key.
        """
        return cls(
            name=json['name'],
            parameters=list(map(lambda p: ContractParameterDefinition.from_json(p), json['parameters']))
        )


class ContractMethodDescriptor(ContractEventDescriptor, IJson):
    """
    A description of a callable method on a contract.
    """

    def __init__(self, name: str,
                 parameters: List[ContractParameterDefinition],
                 return_type: contracts.ContractParameterType):
        """
        Args:
            name: the human readable identifier of the method.
            parameters: the list of parameters the method takes.
            return_type: the type of the returned value.
        """
        super(ContractMethodDescriptor, self).__init__(name, parameters)
        self.return_type = return_type

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (self.name == other.name
                and self.parameters == other.parameters
                and self.return_type == other.return_type)

    @classmethod
    def default_entrypoint(cls) -> ContractMethodDescriptor:
        """
        Create an entry point accepting an operation and a list of arguments.

        Note: deprecated post neo3-preview2.
        """
        warnings.warn("Will be deprecated post neo3-preview2", PendingDeprecationWarning)
        return cls(
            "Main",
            [
                ContractParameterDefinition("operation", contracts.ContractParameterType.STRING),
                ContractParameterDefinition("args", contracts.ContractParameterType.ARRAY)
            ],
            contracts.ContractParameterType.ANY
        )

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        json = super(ContractMethodDescriptor, self).to_json()
        json.update({
            "returnType": self.return_type.name.title()
        })
        return json

    @classmethod
    def from_json(cls, json: dict) -> ContractMethodDescriptor:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary keys.
        """
        return cls(
            name=json['name'],
            parameters=list(map(lambda p: contracts.ContractParameterDefinition.from_json(p), json['parameters'])),
            return_type=contracts.ContractParameterType[json['returnType'].upper()]
        )


class ContractABI(IJson):
    """
    The smart contract application binary interface describes the callable events and contracts for a given
    smart contract.
    """

    def __init__(self,
                 contract_hash: types.UInt160,
                 entry_point: contracts.ContractMethodDescriptor,
                 methods: List[contracts.ContractMethodDescriptor],
                 events: List[contracts.ContractEventDescriptor]):
        """

        Args:
            contract_hash: the result of performing RIPEMD160(SHA256(vm_script)), where vm_script is the smart contract
            byte code.
            entry_point: Obsolete after preview 2.
            methods: the available methods in the contract.
            events: the various events that can be broad casted by the contract.
        """
        self.contract_hash = contract_hash
        self.entry_point = entry_point
        self.methods = methods
        self.events = events

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (self.contract_hash == other.contract_hash
                and self.entry_point == other.entry_point
                and self.methods == other.methods
                and self.events == other.events)

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        json = {
            "hash": '0x' + str(self.contract_hash),
            "entryPoint": self.entry_point.to_json(),
            "methods": list(map(lambda m: m.to_json(), self.methods)),
            "events": list(map(lambda e: e.to_json(), self.events))
        }
        return json

    @classmethod
    def from_json(cls, json: dict) -> ContractABI:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary keys.
        """
        return cls(
            contract_hash=types.UInt160.from_string(json['hash'][2:]),
            entry_point=contracts.ContractMethodDescriptor.from_json(json['entryPoint']),
            methods=list(map(lambda m: contracts.ContractMethodDescriptor.from_json(m), json['methods'])),
            events=list(map(lambda e: contracts.ContractEventDescriptor.from_json(e), json['events'])),
        )



from typing import Any, Dict, Iterable, List, Sized, Tuple

from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.expression import IExpression
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class StorageDeleteMethod(InteropMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        identifier = 'delete'
        syscall = 'System.Storage.Delete'
        self._storage_context = 'System.Storage.GetContext'  # TODO: refactor when default arguments are implemented
        args: Dict[str, Variable] = {'key': Variable(Type.bytes)}
        super().__init__(identifier, syscall, args, return_type=Type.none)

    @property
    def requires_storage(self) -> bool:
        return True

    def validate_parameters(self, *params: IExpression) -> bool:
        if any(not isinstance(param, IExpression) for param in params):
            return False

        args: List[IType] = [arg.type for arg in self.args.values()]
        # TODO: refactor when default arguments are implemented
        if len(params) != len(args):
            return False

        return self._validate_key_type(params[0].type)

    def _validate_key_type(self, key_type: IType):
        # TODO: refactor when `Union` type is implemented
        from boa3.model.type.type import Type
        return Type.str.is_type_of(key_type) or Type.bytes.is_type_of(key_type)

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        opcodes = [(Opcode.SYSCALL, self.storage_context_hash)]
        opcodes.extend(super().opcode)
        return opcodes

    @property
    def storage_context_hash(self) -> bytes:
        # TODO: refactor when default arguments are implemented
        return self._method_hash(self._storage_context)

    @property
    def key_arg(self) -> Variable:
        return self.args['key']

    def build(self, value: Any):
        exp: List[IExpression] = []
        if isinstance(value, Sized):
            if len(value) > 1 or not isinstance(value, Iterable):
                return self
            exp = [exp if isinstance(exp, IExpression) else Variable(exp)
                   for exp in value if isinstance(exp, (IExpression, IType))]

        elif isinstance(exp, (IExpression, IType)):
            exp = [value if isinstance(value, IExpression) else Variable(value)]
        else:
            return self

        if not self.validate_parameters(*exp):
            return self

        method = self
        key_type: IType = exp[0].type
        if not method.key_arg.type.is_type_of(key_type):
            method = StorageDeleteMethod()
            method.args['key'] = Variable(key_type)
        return method



from unittest import TestCase

from boa3.constants import ENCODING
from boa3.neo import to_script_hash
from boa3.neo.contracts.neffile import NefFile
from boa3.neo.vm.type.Integer import Integer


class TestNefFile(TestCase):
    test_version = '1.2.3'
    test_version_bytes = (
        Integer(1).to_byte_array(min_length=4)
        + Integer(2).to_byte_array(min_length=4)
        + Integer(3).to_byte_array(min_length=4)
        + Integer(0).to_byte_array(min_length=4)
    )
    test_script = b'\x01\x02\x03'

    def create_test_nef(self, test_script):
        nef = NefFile(test_script)
        nef._set_version(self.test_version)
        return nef

    def test_empty_script(self):
        script = bytes()
        nef = self.create_test_nef(script)
        self.assertEqual(len(nef._nef.script), 0)

    def test_script_hash(self):
        script = self.test_script
        nef = self.create_test_nef(script)
        self.assertEqual(len(nef.script_hash), 20)
        self.assertEqual(nef.script_hash, to_script_hash(script))

    def test_serialize(self):
        script = self.test_script
        nef = self.create_test_nef(script)
        result = nef.serialize()

        self.assertEqual(len(result), len(nef._nef))

        # the first 4 bytes of the header are from nef magic
        # the next 32 bytes of the header are from the compiler id
        start, end = (4, 36)
        self.assertEqual(result[start:end], nef._nef.compiler.encode(ENCODING))

        # the next 16 bytes are from the version, 4 bytes for each field
        start, end = (end, end + 16)
        self.assertEqual(result[start:end], self.test_version_bytes)

        # the next 20 bytes are from the scripthash of the smart contract
        start, end = (end, end + 20)
        self.assertEqual(result[start:end], nef._nef.script_hash.to_array())

        # the next 4 bytes are from the check sum of the nef file
        start, end = (end, end + 4)
        self.assertEqual(result[start:end], nef._nef.checksum)

        # the next byte is from the size of the smart contract
        # and the last bytes are from the script of the smart contract, up to 1MB
        self.assertEqual(result[end], len(nef._nef.script))
        start, end = (end + 1, end + 1 + 1024 ** 2)
        self.assertEqual(result[start:end], nef._nef.script)



from typing import List, Optional

from boa3.model.operation.binary.arithmetic.addition import Addition
from boa3.model.operation.binary.arithmetic.concat import Concat
from boa3.model.operation.binary.arithmetic.division import Division
from boa3.model.operation.binary.arithmetic.floordivision import FloorDivision
from boa3.model.operation.binary.arithmetic.modulo import Modulo
from boa3.model.operation.binary.arithmetic.multiplication import Multiplication
from boa3.model.operation.binary.arithmetic.power import Power
from boa3.model.operation.binary.arithmetic.strmultiplication import StrMultiplication
from boa3.model.operation.binary.arithmetic.subtraction import Subtraction
from boa3.model.operation.binary.binaryoperation import BinaryOperation
from boa3.model.operation.binary.logical.booleanand import BooleanAnd
from boa3.model.operation.binary.logical.booleanor import BooleanOr
from boa3.model.operation.binary.logical.leftshift import LeftShift
from boa3.model.operation.binary.logical.logicand import LogicAnd
from boa3.model.operation.binary.logical.logicor import LogicOr
from boa3.model.operation.binary.logical.logicxor import LogicXor
from boa3.model.operation.binary.logical.rightshift import RightShift
from boa3.model.operation.binary.relational.LessThan import LessThan
from boa3.model.operation.binary.relational.Lessthanorequal import LessThanOrEqual
from boa3.model.operation.binary.relational.greaterthan import GreaterThan
from boa3.model.operation.binary.relational.greaterthanorequal import GreaterThanOrEqual
from boa3.model.operation.binary.relational.identity import Identity
from boa3.model.operation.binary.relational.notidentity import NotIdentity
from boa3.model.operation.binary.relational.numericequality import NumericEquality
from boa3.model.operation.binary.relational.numericinequality import NumericInequality
from boa3.model.operation.binary.relational.objectequality import ObjectEquality
from boa3.model.operation.binary.relational.objectinequality import ObjectInequality
from boa3.model.operation.operation import IOperation
from boa3.model.operation.operator import Operator
from boa3.model.operation.unary.noneidentity import NoneIdentity
from boa3.model.operation.unary.nonenotidentity import NoneNotIdentity
from boa3.model.type.itype import IType


class BinaryOp:
    # Arithmetic operations
    Add = Addition()
    Sub = Subtraction()
    Mul = Multiplication()
    Div = Division()
    IntDiv = FloorDivision()
    Mod = Modulo()
    Pow = Power()
    Concat = Concat()
    StrMul = StrMultiplication()

    # Relational operations
    NumEq = NumericEquality()
    NumNotEq = NumericInequality()
    Lt = LessThan()
    LtE = LessThanOrEqual()
    Gt = GreaterThan()
    GtE = GreaterThanOrEqual()
    IsNone = NoneIdentity()
    IsNotNone = NoneNotIdentity()
    Is = Identity()
    IsNot = NotIdentity()
    Eq = ObjectEquality()
    NotEq = ObjectInequality()

    # Logical operations
    And = BooleanAnd()
    Or = BooleanOr()
    BitAnd = LogicAnd()
    BitOr = LogicOr()
    Xor = LogicXor()
    LShift = LeftShift()
    RShift = RightShift()

    @classmethod
    def validate_type(cls, operator: Operator, left: IType, right: IType) -> Optional[BinaryOperation]:
        """
        Gets a binary operation given the operator and the operands types.

        :param operator: binary operator
        :param left: type of the left operand
        :param right: type of the right operand
        :return: The operation if exists. None otherwise;
        :rtype: BinaryOperation or None
        """
        for id, op in vars(cls).items():
            if isinstance(op, IOperation) and op.is_valid(operator, left, right):
                if isinstance(op, BinaryOperation):
                    return op.build(left, right)
                else:
                    from boa3.model.type.type import Type
                    operand = right if left is Type.none else left
                    return op.build(operand)

    @classmethod
    def get_operation_by_operator(cls, operator: Operator, left_operand: IType) -> Optional[BinaryOperation]:
        """
        Gets a binary operation given the operator.

        :param operator: binary operator
        :param left_operand: left operand of the operator
        :return: The operation if exists. If exists more than one operation with the same operator, returns the one with
        the same left operand. If none has the same left operand, returns the first found. None otherwise;
        :rtype: BinaryOperation or None
        """
        valid_operations: List[BinaryOperation] = []
        for id, op in vars(cls).items():
            if isinstance(op, BinaryOperation) and op.operator is operator:
                if left_operand in op._valid_types:
                    return op.build(left_operand, op.right_type)
                else:
                    valid_operations.append(op)

        return valid_operations[0] if len(valid_operations) > 0 else None

    @classmethod
    def get_operation(cls, operation: BinaryOperation) -> Optional[BinaryOperation]:
        """
        Gets a binary operation given another operation.

        :param operation: binary operation
        :return: The operation if exists. None otherwise;
        :rtype: BinaryOperation or None
        """
        for id, op in vars(cls).items():
            if type(operation) == type(op):
                return op



from typing import Any

from boa3.model.type.primitive.primitivetype import PrimitiveType
from boa3.neo.vm.type.AbiType import AbiType
from boa3.neo.vm.type.StackItemType import StackItemType


class IntType(PrimitiveType):
    """
    A class used to represent Python int type
    """

    def __init__(self):
        identifier = 'int'
        super().__init__(identifier)

    @property
    def default_value(self) -> Any:
        return int()

    @property
    def abi_type(self) -> AbiType:
        return AbiType.Integer

    @property
    def stack_item(self) -> StackItemType:
        return StackItemType.Integer

    @classmethod
    def build(cls, value: Any):
        if cls._is_type_of(value):
            from boa3.model.type.type import Type
            return Type.int

    @classmethod
    def _is_type_of(cls, value: Any):
        return type(value) is int or isinstance(value, IntType)



from .abi import (ContractParameterType,
                  ContractMethodDescriptor,
                  ContractEventDescriptor,
                  ContractParameterDefinition,
                  ContractABI)
from .contracttypes import (TriggerType)
from .descriptor import (ContractPermissionDescriptor)
from .manifest import (ContractGroup,
                       ContractFeatures,
                       ContractManifest,
                       ContractPermission,
                       WildcardContainer)
from .nef import (NEF, Version)

__all__ = ['ContractParameterType',
           'TriggerType',
           'ContractMethodDescriptor',
           'ContractEventDescriptor',
           'ContractParameterDefinition']



from typing import Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.primitive.primitivetype import PrimitiveType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class ScriptHashMethod(IBuiltinMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        identifier = 'to_script_hash'
        args: Dict[str, Variable] = {'data_bytes': Variable(Type.any)}
        super().__init__(identifier, args, Type.bytes)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1:
            return False
        if not isinstance(params[0], IExpression):
            return False
        return isinstance(params[0].type, PrimitiveType)

    @property
    def is_supported(self) -> bool:
        # TODO: change when implement variable values conversion to script hash
        return False

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        return []

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None



from boa3.boa3 import Boa3
from boa3.constants import ENCODING
from boa3.exception.CompilerError import MismatchedTypes, MissingReturnStatement, TooManyReturns, TypeHintMissing
from boa3.model.type.type import Type
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3_test.tests.boa_test import BoaTest


class TestFunction(BoaTest):

    def test_integer_function(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'           # num local variables
            + b'\x01'           # num arguments
            + Opcode.PUSH10     # body
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/function_test/IntegerFunction.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_string_function(self):
        expected_output = (
            # functions without arguments and local variables don't need initslot
            Opcode.PUSHDATA1        # body
            + bytes([len('42')])
            + bytes('42', ENCODING)
            + Opcode.RET            # return
        )

        path = '%s/boa3_test/test_sc/function_test/StringFunction.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_bool_function(self):
        expected_output = (
            # functions without arguments and local variables don't need initslot
            Opcode.PUSH1      # body
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/function_test/BoolFunction.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_none_function(self):
        path = '%s/boa3_test/test_sc/function_test/NoneFunction.py' % self.dirname

        with self.assertRaises(NotImplementedError):
            output = Boa3.compile(path)

    def test_arg_without_type_hint(self):
        path = '%s/boa3_test/test_sc/function_test/ArgWithoutTypeHintFunction.py' % self.dirname
        self.assertCompilerLogs(TypeHintMissing, path)

    def test_no_return_hint_function_with_empty_return_statement(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'           # num local variables
            + b'\x01'           # num arguments
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/function_test/EmptyReturnFunction.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_no_return_hint_function_without_return_statement(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'           # num local variables
            + b'\x01'           # num arguments
            + Opcode.PUSHNULL
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/function_test/NoReturnFunction.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_return_type_hint_function_with_empty_return(self):
        path = '%s/boa3_test/test_sc/function_test/ExpectingReturnFunction.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_multiple_return_function(self):
        path = '%s/boa3_test/test_sc/function_test/MultipleReturnFunction.py' % self.dirname
        self.assertCompilerLogs(TooManyReturns, path)

    def test_tuple_function(self):
        path = '%s/boa3_test/test_sc/function_test/TupleFunction.py' % self.dirname
        self.assertCompilerLogs(TooManyReturns, path)

    def test_default_return(self):
        path = '%s/boa3_test/test_sc/function_test/DefaultReturn.py' % self.dirname
        self.assertCompilerLogs(MissingReturnStatement, path)

    def test_empty_list_return(self):
        expected_output = (
            Opcode.NEWARRAY0
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/function_test/EmptyListReturn.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_mismatched_return_type(self):
        path = '%s/boa3_test/test_sc/function_test/MismatchedReturnType.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_mismatched_return_type_with_if(self):
        path = '%s/boa3_test/test_sc/function_test/MismatchedReturnTypeWithIf.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_call_void_function_without_args(self):
        called_function_address = Integer(4).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT     # Main
            + b'\x00'
            + b'\x02'
            + Opcode.CALL           # TestFunction()
            + called_function_address
            + Opcode.PUSH1          # return True
            + Opcode.RET
            + Opcode.INITSLOT   # TestFunction
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH1          # a = 1
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET            # return
        )

        path = '%s/boa3_test/test_sc/function_test/CallVoidFunctionWithoutArgs.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_call_function_without_args(self):
        called_function_address = Integer(5).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT     # Main
            + b'\x01'
            + b'\x02'
            + Opcode.CALL           # a = TestFunction()
            + called_function_address
            + Opcode.STLOC0
            + Opcode.LDLOC0         # return a
            + Opcode.RET
            + Opcode.PUSH1      # TestFunction
            + Opcode.RET            # return 1
        )

        path = '%s/boa3_test/test_sc/function_test/CallReturnFunctionWithoutArgs.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_call_void_function_with_literal_args(self):
        called_function_address = Integer(4).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT     # Main
            + b'\x00'
            + b'\x02'
            + Opcode.PUSH2          # TestAdd(1, 2)
            + Opcode.PUSH1
            + Opcode.CALL
            + called_function_address
            + Opcode.PUSH1          # return True
            + Opcode.RET
            + Opcode.INITSLOT   # TestFunction
            + b'\x01'
            + b'\x02'
            + Opcode.LDARG0         # c = a + b
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET            # return
        )

        path = '%s/boa3_test/test_sc/function_test/CallVoidFunctionWithLiteralArgs.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_call_function_with_literal_args(self):
        called_function_address = Integer(5).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT     # Main
            + b'\x01'
            + b'\x02'
            + Opcode.PUSH2          # a = TestAdd(1, 2)
            + Opcode.PUSH1
            + Opcode.CALL
            + called_function_address
            + Opcode.STLOC0
            + Opcode.LDLOC0         # return a
            + Opcode.RET
            + Opcode.INITSLOT   # TestFunction
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0         # return a + b
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.RET            # return
        )

        path = '%s/boa3_test/test_sc/function_test/CallReturnFunctionWithLiteralArgs.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_call_void_function_with_variable_args(self):
        called_function_address = Integer(4).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT     # Main
            + b'\x02'
            + b'\x02'
            + Opcode.PUSH1          # a = 1
            + Opcode.STLOC0
            + Opcode.PUSH2          # b = 2
            + Opcode.STLOC1
            + Opcode.LDLOC1         # TestAdd(a, b)
            + Opcode.LDLOC0
            + Opcode.CALL
            + called_function_address
            + Opcode.PUSH1          # return True
            + Opcode.RET
            + Opcode.INITSLOT   # TestFunction
            + b'\x01'
            + b'\x02'
            + Opcode.LDARG0         # c = a + b
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.PUSHNULL
            + Opcode.RET            # return
        )

        path = '%s/boa3_test/test_sc/function_test/CallVoidFunctionWithVariableArgs.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_call_function_with_variable_args(self):
        called_function_address = Integer(5).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT     # Main
            + b'\x03'
            + b'\x02'
            + Opcode.PUSH1          # a = 1
            + Opcode.STLOC0
            + Opcode.PUSH2          # b = 2
            + Opcode.STLOC1
            + Opcode.LDLOC1         # c = TestAdd(a, b)
            + Opcode.LDLOC0
            + Opcode.CALL
            + called_function_address
            + Opcode.STLOC2
            + Opcode.LDLOC2         # return c
            + Opcode.RET
            + Opcode.INITSLOT   # TestFunction
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0         # return a + b
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.RET            # return
        )

        path = '%s/boa3_test/test_sc/function_test/CallReturnFunctionWithVariableArgs.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_call_function_on_return(self):
        called_function_address = Integer(3).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT     # Main
            + b'\x02'
            + b'\x02'
            + Opcode.PUSH1          # a = 1
            + Opcode.STLOC0
            + Opcode.PUSH2          # b = 2
            + Opcode.STLOC1
            + Opcode.LDLOC1         # return TestAdd(a, b)
            + Opcode.LDLOC0
            + Opcode.CALL
            + called_function_address
            + Opcode.RET
            + Opcode.INITSLOT   # TestFunction
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0         # return a + b
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.RET            # return
        )

        path = '%s/boa3_test/test_sc/function_test/CallReturnFunctionOnReturn.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_call_function_without_variables(self):
        main_to_one_address = Integer(-10).to_byte_array(min_length=1, signed=True)
        main_to_two_address = Integer(5).to_byte_array(min_length=1, signed=True)
        two_to_one_address = Integer(-24).to_byte_array(min_length=1, signed=True)
        end_if = Integer(5).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.PUSH1        # One
            + Opcode.RET            # return 1
            + Opcode.INITSLOT   # Main
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0         # if arg0 == 1
            + Opcode.PUSH1
            + Opcode.NUMEQUAL
            + Opcode.JMPIFNOT
            + end_if
                + Opcode.CALL           # return One()
                + main_to_one_address
                + Opcode.RET
            + Opcode.LDARG0         # elif arg0 == 2
            + Opcode.PUSH2
            + Opcode.NUMEQUAL
            + Opcode.JMPIFNOT
            + end_if
                + Opcode.CALL           # return Two()
                + main_to_two_address
                + Opcode.RET
            + Opcode.PUSH0          # default return
            + Opcode.RET
            + Opcode.PUSH1     # Two
            + Opcode.CALL           # return 1 + One()
            + two_to_one_address
            + Opcode.ADD
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/function_test/CallFunctionWithoutVariables.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_call_function_written_before_caller(self):
        call_address = Integer(-12).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT     # TestFunction
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0         # return a + b
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.RET
            + Opcode.INITSLOT   # Main
            + b'\x00'
            + b'\x02'
            + Opcode.PUSH2          # return TestAdd(a, b)
            + Opcode.PUSH1
            + Opcode.CALL
            + call_address
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/function_test/CallFunctionWrittenBefore.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_return_inside_if(self):
        expected_output = (
            Opcode.INITSLOT     # Main
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # if arg0 % 3 == 1
            + Opcode.PUSH3
            + Opcode.MOD
            + Opcode.PUSH1
            + Opcode.NUMEQUAL
            + Opcode.JMPIFNOT
            + Integer(6).to_byte_array(min_length=1, signed=True)
                + Opcode.LDARG0     # return arg0 - 1
                + Opcode.PUSH1
                + Opcode.SUB
                + Opcode.RET
            + Opcode.LDARG0     # elif arg0 % 3 == 2
            + Opcode.PUSH3
            + Opcode.MOD
            + Opcode.PUSH2
            + Opcode.NUMEQUAL
            + Opcode.JMPIFNOT
            + Integer(6).to_byte_array(min_length=1, signed=True)
                + Opcode.LDARG0     # return arg0 + 1
                + Opcode.PUSH1
                + Opcode.ADD
                + Opcode.RET
            + Opcode.LDARG0     # else
            + Opcode.RET            # return arg0
        )

        path = '%s/boa3_test/test_sc/function_test/ReturnIf.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_missing_return_inside_if(self):
        path = '%s/boa3_test/test_sc/function_test/ReturnIfMissing.py' % self.dirname
        self.assertCompilerLogs(MissingReturnStatement, path)

    def test_missing_return_inside_elif(self):
        path = '%s/boa3_test/test_sc/function_test/ReturnElifMissing.py' % self.dirname
        self.assertCompilerLogs(MissingReturnStatement, path)

    def test_missing_return_inside_else(self):
        path = '%s/boa3_test/test_sc/function_test/ReturnElseMissing.py' % self.dirname
        self.assertCompilerLogs(MissingReturnStatement, path)

    def test_return_inside_multiple_inner_if(self):
        expected_output = (
            Opcode.INITSLOT     # Main
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # if condition
            + Opcode.JMPIFNOT
            + Integer(29).to_byte_array(min_length=1, signed=True)
                + Opcode.LDARG0     # if condition
                + Opcode.JMPIFNOT
                + Integer(14).to_byte_array(min_length=1, signed=True)
                    + Opcode.LDARG0     # if condition
                    + Opcode.JMPIFNOT
                    + Integer(4).to_byte_array(min_length=1, signed=True)
                    + Opcode.PUSH1          # return 1
                    + Opcode.RET
                    + Opcode.LDARG0     # if condition
                    + Opcode.JMPIFNOT
                    + Integer(4).to_byte_array(min_length=1, signed=True)
                    + Opcode.PUSH2          # return 2
                    + Opcode.RET
                    + Opcode.PUSH3      # else
                    + Opcode.RET            # return 3
                + Opcode.LDARG0     # elif condition
                + Opcode.JMPIFNOT
                + Integer(9).to_byte_array(min_length=1, signed=True)
                    + Opcode.LDARG0     # if condition
                    + Opcode.JMPIFNOT
                    + Integer(4).to_byte_array(min_length=1, signed=True)
                    + Opcode.PUSH4          # return 4
                    + Opcode.RET
                    + Opcode.PUSH5      # else
                    + Opcode.RET            # return 5
                + Opcode.PUSH6      # else
                + Opcode.RET            # return 6
            + Opcode.LDARG0     # else
            + Opcode.JMPIFNOT       # if condition
            + Integer(4).to_byte_array(min_length=1, signed=True)
                + Opcode.PUSH7          # return 7
                + Opcode.RET
            + Opcode.LDARG0         # if condition
            + Opcode.JMPIFNOT
            + Integer(4).to_byte_array(min_length=1, signed=True)
                + Opcode.PUSH8          # return 8
                + Opcode.RET
            + Opcode.PUSH9          # else
            + Opcode.RET                # return 9
        )

        path = '%s/boa3_test/test_sc/function_test/ReturnMultipleInnerIf.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_missing_return_inside_multiple_inner_if(self):
        path = '%s/boa3_test/test_sc/function_test/ReturnMultipleInnerIfMissing.py' % self.dirname
        self.assertCompilerLogs(MissingReturnStatement, path)

    def test_return_if_expression(self):
        expected_output = (
            Opcode.INITSLOT     # Main
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # return 5 if condition else 10
            + Opcode.JMPIFNOT
                + Integer(5).to_byte_array(min_length=1, signed=True)
                + Opcode.PUSH5      # 5
            + Opcode.JMP        # else
            + Integer(3).to_byte_array(min_length=1, signed=True)
                + Opcode.PUSH10     # 10
            + Opcode.RET        # return
        )

        path = '%s/boa3_test/test_sc/function_test/ReturnIfExpression.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_return_if_expression_mismatched_type(self):
        path = '%s/boa3_test/test_sc/function_test/ReturnIfExpressionMismatched.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_return_inside_for(self):
        expected_output = (
            Opcode.INITSLOT     # Main
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0     # for_sequence = arg0
            + Opcode.PUSH0      # for_index = 0
            + Opcode.JMP        # begin for
            + Integer(17).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER     # value = for_sequence[for_index]
                + Opcode.OVER
                    + Opcode.DUP
                    + Opcode.SIGN
                    + Opcode.PUSHM1
                    + Opcode.JMPNE
                    + Integer(5).to_byte_array(min_length=1, signed=True)
                    + Opcode.OVER
                    + Opcode.SIZE
                    + Opcode.ADD
                + Opcode.PICKITEM
                + Opcode.STLOC0
                + Opcode.LDLOC0     # return value
                + Opcode.RET
                + Opcode.INC     # for_index = for_index + 1
            + Opcode.DUP        # if for_index < len(for_sequence)
            + Opcode.PUSH2
            + Opcode.PICK
            + Opcode.SIZE
            + Opcode.LT
            + Opcode.JMPIF
            + Integer(-20).to_byte_array(min_length=1, signed=True)
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.PUSH5      # else
            + Opcode.RET          # return 5
        )

        path = '%s/boa3_test/test_sc/function_test/ReturnFor.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_missing_return_inside_for(self):
        expected_output = (
            Opcode.INITSLOT     # Main
            + b'\x02'
            + b'\x01'
            + Opcode.PUSH0      # x = 0
            + Opcode.STLOC0
            + Opcode.LDARG0     # for_sequence = arg0
            + Opcode.PUSH0      # for_index = 0
            + Opcode.JMP        # begin for
            + Integer(19).to_byte_array(min_length=1, signed=True)
                + Opcode.OVER       # value = for_sequence[for_index]
                + Opcode.OVER
                    + Opcode.DUP
                    + Opcode.SIGN
                    + Opcode.PUSHM1
                    + Opcode.JMPNE
                    + Integer(5).to_byte_array(min_length=1, signed=True)
                    + Opcode.OVER
                    + Opcode.SIZE
                    + Opcode.ADD
                + Opcode.PICKITEM
                + Opcode.STLOC1
                + Opcode.LDLOC0     # x += value
                + Opcode.LDLOC1
                + Opcode.ADD
                + Opcode.STLOC0
                + Opcode.INC        # for_index = for_index + 1
            + Opcode.DUP        # if for_index < len(for_sequence)
            + Opcode.PUSH2
            + Opcode.PICK
            + Opcode.SIZE
            + Opcode.LT
            + Opcode.JMPIF
            + Integer(-22).to_byte_array(min_length=1, signed=True)
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.LDLOC0     # else
            + Opcode.RET          # return x
        )

        path = '%s/boa3_test/test_sc/function_test/ReturnForOnlyOnElse.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_missing_return_inside_for_else(self):
        path = '%s/boa3_test/test_sc/function_test/ReturnForElseMissing.py' % self.dirname
        self.assertCompilerLogs(MissingReturnStatement, path)

    def test_return_inside_while(self):
        expected_output = (
            Opcode.INITSLOT     # Main
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0     # x = arg0
            + Opcode.STLOC0
            + Opcode.JMP        # begin while
            + Integer(8).to_byte_array(min_length=1, signed=True)
                + Opcode.LDLOC0     # x += 1
                + Opcode.PUSH1
                + Opcode.ADD
                + Opcode.STLOC0
                + Opcode.LDLOC0     # return x
                + Opcode.RET
            + Opcode.LDLOC0
            + Opcode.PUSH10
            + Opcode.LT
            + Opcode.JMPIF      # end while x < 10
            + Integer(-9).to_byte_array(min_length=1, signed=True)
            + Opcode.LDLOC0     # else
            + Opcode.RET            # return x
        )

        path = '%s/boa3_test/test_sc/function_test/ReturnWhile.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_missing_return_inside_while(self):
        expected_output = (
            Opcode.INITSLOT     # Main
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0     # x = arg0
            + Opcode.STLOC0
            + Opcode.JMP        # begin while
            + Integer(6).to_byte_array(min_length=1, signed=True)
                + Opcode.LDLOC0     # x += 1
                + Opcode.PUSH1
                + Opcode.ADD
                + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.PUSH10
            + Opcode.LT
            + Opcode.JMPIF      # end while x < 10
            + Integer(-7).to_byte_array(min_length=1, signed=True)
            + Opcode.LDLOC0     # else
            + Opcode.RET            # return x
        )

        path = '%s/boa3_test/test_sc/function_test/ReturnWhileOnlyOnElse.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_missing_return_inside_while_without_else(self):
        path = '%s/boa3_test/test_sc/function_test/ReturnWhileWithoutElse.py' % self.dirname
        self.assertCompilerLogs(MissingReturnStatement, path)

    def test_multiple_function_large_call(self):
        expected_output = (
            # main(operation: str, arg: List[int])
            Opcode.INITSLOT + b'\x02\x02'
            + Opcode.LDARG0         # if operation == 'calculate' and len(arg) >= 2:
            + Opcode.PUSHDATA1 + b'\x09calculate'
            + Opcode.EQUAL
            + Opcode.LDARG1
            + Opcode.SIZE
            + Opcode.PUSH2
            + Opcode.GE
            + Opcode.BOOLAND
            + Opcode.JMPIFNOT + b'\x39'
            + Opcode.NEWARRAY0          # operands: List[int] = []
            + Opcode.STLOC0
            + Opcode.PUSH1              # i = 1
            + Opcode.STLOC1
            + Opcode.JMP + b'\x1e'      # begin while i < len(arg):
            + Opcode.LDLOC0                 # operands.append(arg[i])
            + Opcode.LDARG1                 # arg[i]
            + Opcode.LDLOC1
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE + b'\x05'
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM               # append(args[i])
                + Opcode.OVER
                + Opcode.ISTYPE
                + Type.bytearray.stack_item
                + Opcode.JMPIFNOT
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.CAT
                + Opcode.JMP
                + Integer(5).to_byte_array(min_length=1)
                + Opcode.APPEND
                + Opcode.JMP
                + Integer(3).to_byte_array(min_length=1)
                + Opcode.STLOC0
            + Opcode.LDLOC1                 # i += 1
            + Opcode.PUSH1
            + Opcode.ADD
            + Opcode.STLOC1
            + Opcode.LDLOC1             # end while i < len(arg)
            + Opcode.LDARG1
            + Opcode.SIZE
            + Opcode.LT
            + Opcode.JMPIF + b'\xe0'
            + Opcode.LDLOC0             # return calculate(arg[0], operands)
            + Opcode.LDARG1             # arg[0]
            + Opcode.PUSH0
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE + b'\x05'
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.CALL + b'\x05'     # calculate(arg[0], operands)
            + Opcode.RET
            + Opcode.PUSHNULL           # else: return None
            + Opcode.RET
            # calculate(op_id: int, operands: List[int])
            + Opcode.INITSLOT + b'\x05\x02'
            + Opcode.LDARG0             # op = get_operation(op_id)
            + Opcode.CALL + b'\x62'
            + Opcode.STLOC0
            + Opcode.LDARG1             # if len(operands) <= 0
            + Opcode.SIZE
            + Opcode.PUSH0
            + Opcode.LE
            + Opcode.JMPIFNOT + b'\x2b'
            + Opcode.PUSHDATA1 + b'\x24There are missing some parameters...'
            + Opcode.STLOC1                 # result = "There are missing some parameters..."
            + Opcode.JMP + b'\x30'      # else
            + Opcode.LDARG1                 # calc = operands[0]
            + Opcode.PUSH0
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE + b'\x05'
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.STLOC2
            + Opcode.PUSH1                  # y = 1
            + Opcode.STLOC3
            + Opcode.LDARG1                 # size = len(operands)
            + Opcode.SIZE
            + Opcode.STLOC4
            + Opcode.JMP + b'\x16'          # begin while
            + Opcode.LDARG1                     # operands[y]
            + Opcode.LDLOC3
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE + b'\x05'
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.LDLOC2
            + Opcode.LDLOC0
            + Opcode.CALL + b'\x6e'             # calc = calculate_simple(op, calc, operands[y])
            + Opcode.STLOC2
            + Opcode.LDLOC3                     # y += 1
            + Opcode.PUSH1
            + Opcode.ADD
            + Opcode.STLOC3
            + Opcode.LDLOC3                 # end while y < size
            + Opcode.LDLOC4
            + Opcode.LT
            + Opcode.JMPIF + b'\xe9'
            + Opcode.LDLOC2                 # else: result = calc
            + Opcode.STLOC1
            + Opcode.LDLOC1                 # return result
            + Opcode.RET
            # def get_operation(args: int)
            + Opcode.INITSLOT + b'\x00\x01'
            + Opcode.LDARG0                 # if args == 1:
            + Opcode.PUSH1
            + Opcode.NUMEQUAL
            + Opcode.JMPIFNOT + b'\x08'
            + Opcode.PUSHDATA1 + b'\x03add'     # return "add"
            + Opcode.RET
            + Opcode.LDARG0                 # elif args == 2:
            + Opcode.PUSH2
            + Opcode.NUMEQUAL
            + Opcode.JMPIFNOT + b'\x08'
            + Opcode.PUSHDATA1 + b'\x03sub'     # return "sub"
            + Opcode.RET
            + Opcode.LDARG0                 # elif args == 3:
            + Opcode.PUSH3
            + Opcode.NUMEQUAL
            + Opcode.JMPIFNOT + b'\x08'
            + Opcode.PUSHDATA1 + b'\x03div'     # return "div"
            + Opcode.RET
            + Opcode.LDARG0                 # elif args == 4:
            + Opcode.PUSH4
            + Opcode.NUMEQUAL
            + Opcode.JMPIFNOT + b'\x08'
            + Opcode.PUSHDATA1 + b'\x03mul'     # return "mul"
            + Opcode.RET
            + Opcode.LDARG0                 # elif args == 5:
            + Opcode.PUSH5
            + Opcode.NUMEQUAL
            + Opcode.JMPIFNOT + b'\x08'
            + Opcode.PUSHDATA1 + b'\x03mod'     # return "mod"
            + Opcode.RET
            + Opcode.PUSHDATA1 + b'\x21Operation must be between 1 and 5'
            + Opcode.RET                    # else: return "Operation must be between 1 and 5"
            # calculate_simple(operation: str, a: int, b: int)
            + Opcode.INITSLOT + b'\x05\x03'
            + Opcode.PUSHDATA1 + b'\x03add'  # add_op = "add"
            + Opcode.STLOC0
            + Opcode.PUSHDATA1 + b'\x03sub'  # sub_op = "sub"
            + Opcode.STLOC1
            + Opcode.PUSHDATA1 + b'\x03mul'  # mul_op = "mul"
            + Opcode.STLOC2
            + Opcode.PUSHDATA1 + b'\x03div'  # div_op = "div"
            + Opcode.STLOC3
            + Opcode.PUSHDATA1 + b'\x03mod'  # mod_op = "mod"
            + Opcode.STLOC4
            + Opcode.LDARG0                 # if operation == add_op:
            + Opcode.LDLOC0
            + Opcode.EQUAL
            + Opcode.JMPIFNOT + b'\x07'
            + Opcode.LDARG2                     # return add(a, b)
            + Opcode.LDARG1
            + Opcode.CALL + b'\x34'
            + Opcode.RET
            + Opcode.LDARG0                 # elif operation == sub_op:
            + Opcode.LDLOC1
            + Opcode.EQUAL
            + Opcode.JMPIFNOT + b'\x07'
            + Opcode.LDARG2                     # return sub(a, b)
            + Opcode.LDARG1
            + Opcode.CALL + b'\x23'
            + Opcode.RET
            + Opcode.LDARG0                 # elif operation == div_op:
            + Opcode.LDLOC3
            + Opcode.EQUAL
            + Opcode.JMPIFNOT + b'\x07'
            + Opcode.LDARG2                     # return div(a, b)
            + Opcode.LDARG1
            + Opcode.CALL + b"\x27"
            + Opcode.RET
            + Opcode.LDARG0                 # elif operation == mul_op:
            + Opcode.LDLOC2
            + Opcode.EQUAL
            + Opcode.JMPIFNOT + b'\x07'
            + Opcode.LDARG2                     # return mul(a, b)
            + Opcode.LDARG1
            + Opcode.CALL + b'\x24'
            + Opcode.RET
            + Opcode.LDARG0                 # elif operation == mod_op:
            + Opcode.LDLOC4
            + Opcode.EQUAL
            + Opcode.JMPIFNOT + b'\x07'
            + Opcode.LDARG2                     # return mod(a, b)
            + Opcode.LDARG1
            + Opcode.CALL + b'\x21'
            + Opcode.RET
            + Opcode.PUSH0                  # else:
            + Opcode.RET                        # return 0
            # sub(a: int, b: int)
            + Opcode.INITSLOT + b'\x00\x02'
            + Opcode.LDARG0                 # return a - b
            + Opcode.LDARG1
            + Opcode.SUB
            + Opcode.RET
            # add(a: int, b: int)
            + Opcode.INITSLOT + b'\x00\x02'
            + Opcode.LDARG0                 # return a + b
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.RET
            # div(a: int, b: int)
            + Opcode.INITSLOT + b'\x00\x02'
            + Opcode.LDARG0                 # return a // b
            + Opcode.LDARG1
            + Opcode.DIV
            + Opcode.RET
            # mul(a: int, b: int)
            + Opcode.INITSLOT + b'\x00\x02'
            + Opcode.LDARG0                 # return a * b
            + Opcode.LDARG1
            + Opcode.MUL
            + Opcode.RET
            # mod(a: int, b: int)
            + Opcode.INITSLOT + b'\x00\x02'
            + Opcode.LDARG0                 # return a % b
            + Opcode.LDARG1
            + Opcode.MOD
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/function_test/MultipleFunctionLargeCall.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_function_with_default_argument(self):
        expected_output = (
            Opcode.PUSH3    # add(1, 2, 3)
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.CALL
            + Integer(9).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSH0
            + Opcode.PUSH6  # add(5, 6)
            + Opcode.PUSH5
            + Opcode.CALL
            + Integer(4).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSHNULL
            + Opcode.RET
            + Opcode.INITSLOT   # def add(a: int, b: int, c: int = 0)
            + b'\x00\x03'
            + Opcode.LDARG0     # return a + b + c
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.LDARG2
            + Opcode.ADD
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/function_test/FunctionWithDefaultArgument.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_function_with_only_default_arguments(self):
        expected_output = (
            Opcode.PUSH0        # defaults
            + Opcode.PUSH0
            + Opcode.PUSH0
            + Opcode.CALL   # add()
            + Integer(19).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSH0      # defaults
            + Opcode.PUSH6      # add(5, 6)
            + Opcode.PUSH5
            + Opcode.CALL
            + Integer(14).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSH0      # defaults
            + Opcode.PUSH0
            + Opcode.PUSH9      # add(9)
            + Opcode.CALL
            + Integer(9).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSH3      # add(1, 2, 3)
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.CALL
            + Integer(4).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSHNULL
            + Opcode.RET
            + Opcode.INITSLOT   # def add(a: int, b: int, c: int)
            + b'\x00\x03'
            + Opcode.LDARG0     # return a + b + c
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.LDARG2
            + Opcode.ADD
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/function_test/FunctionWithOnlyDefaultArguments.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_function_with_default_argument_between_other_args(self):
        path = '%s/boa3_test/test_sc/function_test/FunctionWithDefaultArgumentBetweenArgs.py' % self.dirname

        with self.assertRaises(SyntaxError):
            output = Boa3.compile(path)

    def test_call_function_with_kwargs(self):
        path = '%s/boa3_test/test_sc/function_test/CallFunctionWithKwargs.py' % self.dirname

        with self.assertRaises(NotImplementedError):
            output = Boa3.compile(path)



from .flightinfo import FlightInfo
from .nodemanager import NodeManager
from .nodeweight import NodeWeight
from .requestinfo import RequestInfo
from .singleton import _Singleton
from .syncmanager import SyncManager

__all__ = ['FlightInfo', 'NodeWeight', 'RequestInfo', '_Singleton', 'NodeManager', 'SyncManager']



from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.collection.sequence.mutable.mutablesequencetype import MutableSequenceType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class AppendMethod(IBuiltinMethod):
    def __init__(self, sequence_type: MutableSequenceType = None):
        if not isinstance(sequence_type, MutableSequenceType):
            from boa3.model.type.type import Type
            self_arg = Variable(Type.mutableSequence)
            item_arg = Variable(Type.any)
        else:
            self_arg = Variable(sequence_type)
            item_arg = Variable(sequence_type.value_type)

        identifier = 'append'
        args: Dict[str, Variable] = {'self': self_arg, 'item': item_arg}
        super().__init__(identifier, args)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    @property
    def stores_on_slot(self) -> bool:
        return True

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 2:
            return False
        if not all(isinstance(param, IExpression) for param in params):
            return False

        from boa3.model.type.itype import IType
        sequence_type: IType = params[0].type
        value_type: IType = params[1].type

        if not isinstance(sequence_type, MutableSequenceType):
            return False
        return sequence_type.value_type.is_type_of(value_type)

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.model.type.type import Type
        from boa3.neo.vm.type.Integer import Integer
        return [
            (Opcode.OVER, b''),
            (Opcode.ISTYPE, Type.bytearray.stack_item),     # append opcode only works for array
            (Opcode.JMPIFNOT, Integer(5).to_byte_array(min_length=1)),  # when it's bytearray, concatenates the value
            (Opcode.CAT, b''),
            (Opcode.JMP, Integer(5).to_byte_array(min_length=1)),
            (Opcode.APPEND, b'')
        ]

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any):
        if type(value) == type(self.args['self'].type):
            return self
        if isinstance(value, MutableSequenceType):
            return AppendMethod(value)
        return super().build(value)



from enum import Enum


class Operator(str, Enum):
    # Arithmetic operators
    Plus = '+'
    Minus = '-'
    Mult = '*'
    Div = '/'
    IntDiv = '//'
    Mod = '%'
    Pow = '**'

    # Relational operators
    Eq = '=='
    NotEq = '!='
    Lt = '<'
    LtE = '<='
    Gt = '>'
    GtE = '>='
    Is = 'is'
    IsNot = 'is not'

    # Logical operators
    And = 'and'
    Or = 'or'
    Not = 'not'
    BitAnd = '&'
    BitOr = '|'
    BitNot = '~'
    BitXor = '^'
    LeftShift = '<<'
    RightShift = '>>'

    # Other operators
    Subscript = '[]'

    def __str__(self) -> str:
        return self.value



from boa3.boa3 import Boa3
from boa3.exception.CompilerError import MetadataInformationMissing, MismatchedTypes
from boa3.model.builtin.interop.interop import Interop
from boa3.model.type.type import Type
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest


class TestStorage(BoaTest):

    def test_storage_get_bytes_key(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.StorageGet.storage_context_hash
            + Opcode.SYSCALL
            + Interop.StorageGet.interop_method_hash
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/storage_test/StorageGetBytesKey.py' % self.dirname
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        self.assertIn('features', manifest)
        self.assertIn('storage', manifest['features'])
        self.assertEqual(True, manifest['features']['storage'])

    def test_storage_get_str_key(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.StorageGet.storage_context_hash
            + Opcode.SYSCALL
            + Interop.StorageGet.interop_method_hash
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/storage_test/StorageGetStrKey.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_storage_get_mismatched_type(self):
        path = '%s/boa3_test/test_sc/storage_test/StorageGetMismatchedType.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_storage_get_without_metadata(self):
        path = '%s/boa3_test/test_sc/storage_test/StorageGetWithoutMetadata.py' % self.dirname
        self.assertCompilerLogs(MetadataInformationMissing, path)

    def test_storage_put_bytes_key_bytes_value(self):
        value = b'\x01\x02\x03'
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.PUSHDATA1
            + Integer(len(value)).to_byte_array(min_length=1, signed=True)
            + value
            + Opcode.CONVERT + Type.bytes.stack_item
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.StoragePut.storage_context_hash
            + Opcode.SYSCALL
            + Interop.StoragePut.interop_method_hash
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/storage_test/StoragePutBytesKeyBytesValue.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_storage_put_bytes_key_int_value(self):
        value = Integer(123).to_byte_array()
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.PUSHDATA1
            + Integer(len(value)).to_byte_array(min_length=1, signed=True)
            + value
            + Opcode.CONVERT + Type.int.stack_item
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.StoragePut.storage_context_hash
            + Opcode.SYSCALL
            + Interop.StoragePut.interop_method_hash
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/storage_test/StoragePutBytesKeyIntValue.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_storage_put_bytes_key_str_value(self):
        value = String('123').to_bytes()
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.PUSHDATA1
            + Integer(len(value)).to_byte_array(min_length=1, signed=True)
            + value
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.StoragePut.storage_context_hash
            + Opcode.SYSCALL
            + Interop.StoragePut.interop_method_hash
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/storage_test/StoragePutBytesKeyStrValue.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_storage_put_str_key_bytes_value(self):
        value = b'\x01\x02\x03'
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.PUSHDATA1
            + Integer(len(value)).to_byte_array(min_length=1, signed=True)
            + value
            + Opcode.CONVERT + Type.bytes.stack_item
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.StoragePut.storage_context_hash
            + Opcode.SYSCALL
            + Interop.StoragePut.interop_method_hash
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/storage_test/StoragePutStrKeyBytesValue.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_storage_put_str_key_int_value(self):
        value = Integer(123).to_byte_array()
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.PUSHDATA1
            + Integer(len(value)).to_byte_array(min_length=1, signed=True)
            + value
            + Opcode.CONVERT + Type.int.stack_item
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.StoragePut.storage_context_hash
            + Opcode.SYSCALL
            + Interop.StoragePut.interop_method_hash
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/storage_test/StoragePutStrKeyIntValue.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_storage_put_str_key_str_value(self):
        value = String('123').to_bytes()
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.PUSHDATA1
            + Integer(len(value)).to_byte_array(min_length=1, signed=True)
            + value
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.StoragePut.storage_context_hash
            + Opcode.SYSCALL
            + Interop.StoragePut.interop_method_hash
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/storage_test/StoragePutStrKeyStrValue.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_storage_put_mismatched_type_key(self):
        path = '%s/boa3_test/test_sc/storage_test/StoragePutMismatchedTypeKey.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_storage_put_mismatched_type_value(self):
        path = '%s/boa3_test/test_sc/storage_test/StoragePutMismatchedTypeValue.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_storage_delete_bytes_key(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.StorageDelete.storage_context_hash
            + Opcode.SYSCALL
            + Interop.StorageDelete.interop_method_hash
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/storage_test/StorageDeleteBytesKey.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_storage_delete_str_key(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.StorageDelete.storage_context_hash
            + Opcode.SYSCALL
            + Interop.StorageDelete.interop_method_hash
            + Opcode.PUSHNULL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/storage_test/StorageDeleteStrKey.py' % self.dirname
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_storage_delete_mismatched_type(self):
        path = '%s/boa3_test/test_sc/storage_test/StorageDeleteMismatchedType.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)



from boa3.boa3 import Boa3
from boa3.exception.CompilerError import MismatchedTypes, NotSupportedOperation
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3_test.tests.boa_test import BoaTest


class TestRelational(BoaTest):

    def test_number_equality_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.NUMEQUAL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/relational_test/NumEquality.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_number_inequality_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.NUMNOTEQUAL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/relational_test/NumInequality.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_number_inequality_operation_2(self):
        path = '%s/boa3_test/test_sc/relational_test/NumInequalityPython2.py' % self.dirname

        with self.assertRaises(SyntaxError):
            output = Boa3.compile(path)

    def test_number_less_than_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.LT
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/relational_test/NumLessThan.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_number_less_or_equal_than_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.LE
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/relational_test/NumLessOrEqual.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_number_greater_than_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.GT
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/relational_test/NumGreaterThan.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_number_greater_or_equal_than_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.GE
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/relational_test/NumGreaterOrEqual.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_identity_operation(self):
        path = '%s/boa3_test/test_sc/relational_test/NumIdentity.py' % self.dirname
        self.assertCompilerLogs(NotSupportedOperation, path)

    def test_boolean_equality_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.NUMEQUAL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/relational_test/BoolEquality.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_boolean_inequality_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.NUMNOTEQUAL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/relational_test/BoolInequality.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_multiple_comparisons(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x03'
            + Opcode.LDARG1
            + Opcode.LDARG0
            + Opcode.LE
            + Opcode.LDARG0
            + Opcode.LDARG2
            + Opcode.LE
            + Opcode.BOOLAND
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/relational_test/NumRange.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_string_equality_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.EQUAL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/relational_test/StrEquality.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_string_inequality_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.NOTEQUAL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/relational_test/StrInequality.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_string_less_than_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.LT
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/relational_test/StrLessThan.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_string_less_or_equal_than_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.LE
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/relational_test/StrLessOrEqual.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_string_greater_than_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.GT
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/relational_test/StrGreaterThan.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_string_greater_or_equal_than_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.GE
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/relational_test/StrGreaterOrEqual.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_mixed_equality_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.EQUAL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/relational_test/MixedEquality.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_mixed_inequality_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.NOTEQUAL
            + Opcode.RET
        )

        path = '%s/boa3_test/test_sc/relational_test/MixedInequality.py' % self.dirname
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_mixed_less_than_operation(self):
        path = '%s/boa3_test/test_sc/relational_test/MixedLessThan.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_mixed_less_or_equal_than_operation(self):
        path = '%s/boa3_test/test_sc/relational_test/MixedLessOrEqual.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_mixed_greater_than_operation(self):
        path = '%s/boa3_test/test_sc/relational_test/MixedGreaterThan.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)

    def test_mixed_greater_or_equal_than_operation(self):
        path = '%s/boa3_test/test_sc/relational_test/MixedGreaterOrEqual.py' % self.dirname
        self.assertCompilerLogs(MismatchedTypes, path)



import os
from typing import Any, Dict, Tuple

from boa3 import constants
from boa3.boa3 import Boa3
from boa3.compiler.compiler import Compiler
from boa3.constants import BYTEORDER, ENCODING
from boa3.model.event import Event
from boa3.model.method import Method
from boa3.neo.contracts.neffile import NefFile
from boa3.neo.vm.type.AbiType import AbiType
from boa3_test.tests.boa_test import BoaTest


class TestFileGeneration(BoaTest):
    def get_output(self, path: str) -> Tuple[bytes, Dict[str, Any]]:
        nef_output = path.replace('.py', '.nef')
        manifest_output = path.replace('.py', '.manifest.json')

        from boa3.neo.contracts.neffile import NefFile

        if not os.path.isfile(nef_output):
            output = bytes()
        else:
            with open(nef_output, mode='rb') as nef:
                file = nef.read()
                output = NefFile.deserialize(file).script

        if not os.path.isfile(manifest_output):
            manifest = {}
        else:
            with open(manifest_output) as manifest_output:
                import json
                manifest = json.loads(manifest_output.read())

        return output, manifest

    def test_generate_files(self):
        path = '%s/boa3_test/test_sc/generation_test/GenerationWithDecorator.py' % self.dirname
        expected_nef_output = path.replace('.py', '.nef')
        expected_manifest_output = path.replace('.py', '.manifest.json')
        expected_debug_info_output = path.replace('.py', '.nefdbgnfo')
        Boa3.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_nef_output))
        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertTrue(os.path.exists(expected_debug_info_output))

    def test_generate_nef_file(self):
        path = '%s/boa3_test/test_sc/generation_test/GenerationWithDecorator.py' % self.dirname
        expected_nef_output = path.replace('.py', '.nef')
        Boa3.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_nef_output))
        with open(expected_nef_output, 'rb') as nef_output:
            magic = nef_output.read(constants.SIZE_OF_INT32)
            compiler = nef_output.read(32)
            version = nef_output.read(16)
            hash = nef_output.read(constants.SIZE_OF_INT160)
            check_sum = nef_output.read(constants.SIZE_OF_INT32)
            script_size = nef_output.read(1)
            script = nef_output.read()

        self.assertEqual(int.from_bytes(script_size, BYTEORDER), len(script))

        nef = NefFile(script)._nef
        self.assertEqual(compiler.decode(ENCODING), nef.compiler)
        self.assertEqual(hash, nef.script_hash.to_array())
        self.assertEqual(check_sum, nef.checksum)
        self.assertEqual(version, nef.version.to_array())

    def test_generate_manifest_file_with_decorator(self):
        path = '%s/boa3_test/test_sc/generation_test/GenerationWithDecorator.py' % self.dirname
        expected_manifest_output = path.replace('.py', '.manifest.json')
        output, manifest = self.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertNotIn('entrypoint', abi)
        self.assertIn('methods', abi)
        self.assertEqual(2, len(abi['methods']))

        # method Main
        method0 = abi['methods'][0]
        self.assertIn('returntype', method0)
        self.assertEqual(AbiType.Integer, method0['returntype'])
        self.assertIn('parameters', method0)
        self.assertEqual(2, len(method0['parameters']))

        arg0 = method0['parameters'][0]
        self.assertIn('name', arg0)
        self.assertEqual('a', arg0['name'])
        self.assertIn('type', arg0)
        self.assertEqual(AbiType.Integer, arg0['type'])

        arg1 = method0['parameters'][1]
        self.assertIn('name', arg1)
        self.assertEqual('b', arg1['name'])
        self.assertIn('type', arg1)
        self.assertEqual(AbiType.Integer, arg1['type'])

        # method Sub
        method1 = abi['methods'][1]
        self.assertIn('returntype', method1)
        self.assertEqual(AbiType.Integer, method1['returntype'])
        self.assertIn('parameters', method1)
        self.assertEqual(2, len(method1['parameters']))

        arg0 = method1['parameters'][0]
        self.assertIn('name', arg0)
        self.assertEqual('a', arg0['name'])
        self.assertIn('type', arg0)
        self.assertEqual(AbiType.Integer, arg0['type'])

        arg1 = method1['parameters'][1]
        self.assertIn('name', arg1)
        self.assertEqual('b', arg1['name'])
        self.assertIn('type', arg1)
        self.assertEqual(AbiType.Integer, arg1['type'])

        self.assertIn('events', abi)
        self.assertEqual(0, len(abi['events']))

    def test_generate_manifest_file_without_decorator(self):
        path = '%s/boa3_test/test_sc/generation_test/GenerationWithoutDecorator.py' % self.dirname
        expected_manifest_output = path.replace('.py', '.manifest.json')
        output, manifest = self.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('methods', abi)
        self.assertEqual(0, len(abi['methods']))

        self.assertIn('events', abi)
        self.assertEqual(0, len(abi['events']))

    def test_generate_manifest_file_with_event(self):
        path = '%s/boa3_test/test_sc/event_test/EventWithArgument.py' % self.dirname
        expected_manifest_output = path.replace('.py', '.manifest.json')
        compiler = Compiler()
        compiler.compile_and_save(path, path.replace('.py', '.nef'))
        events: Dict[str, Event] = {
            name: method
            for name, method in self.get_compiler_analyser(compiler).symbol_table.items()
            if isinstance(method, Event)
        }

        output, manifest = self.get_output(path)
        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('methods', abi)
        self.assertEqual(0, len(abi['methods']))

        self.assertIn('events', abi)
        self.assertEqual(1, len(abi['events']))

        for abi_event in abi['events']:
            self.assertIn('name', abi_event)
            self.assertIn(abi_event['name'], events)
            self.assertIn('parameters', abi_event)

            event_args = events[abi_event['name']].args
            for event_param in abi_event['parameters']:
                self.assertIn('name', event_param)
                self.assertIn(event_param['name'], event_args)
                self.assertIn('type', event_param)
                self.assertEqual(event_args[event_param['name']].type.abi_type,
                                 event_param['type'])

    def test_generate_manifest_file_with_nep5_transfer_event(self):
        path = '%s/boa3_test/test_sc/event_test/EventNep5Transfer.py' % self.dirname
        expected_manifest_output = path.replace('.py', '.manifest.json')
        compiler = Compiler()
        compiler.compile_and_save(path, path.replace('.py', '.nef'))
        events: Dict[str, Event] = {
            event.name: event
            for event in self.get_compiler_analyser(compiler).symbol_table.values()
            if isinstance(event, Event)
        }

        output, manifest = self.get_output(path)
        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('methods', abi)
        self.assertEqual(0, len(abi['methods']))

        self.assertIn('events', abi)
        self.assertEqual(1, len(abi['events']))

        for abi_event in abi['events']:
            self.assertIn('name', abi_event)
            self.assertIn(abi_event['name'], events)
            self.assertIn('parameters', abi_event)

            event_args = events[abi_event['name']].args
            for event_param in abi_event['parameters']:
                self.assertIn('name', event_param)
                self.assertIn(event_param['name'], event_args)
                self.assertIn('type', event_param)
                self.assertEqual(event_args[event_param['name']].type.abi_type,
                                 event_param['type'])

    def test_generate_nefdbgnfo_file(self):
        path = '%s/boa3_test/test_sc/generation_test/GenerationWithDecorator.py' % self.dirname

        expected_nef_output = path.replace('.py', '.nefdbgnfo')
        compiler = Compiler()
        compiler.compile_and_save(path, path.replace('.py', '.nef'))
        methods: Dict[str, Method] = {
            name: method
            for name, method in self.get_compiler_analyser(compiler).symbol_table.items()
            if isinstance(method, Method)
        }

        self.assertTrue(os.path.exists(expected_nef_output))
        debug_info = self.get_debug_info(path)
        self.assertNotIn('entrypoint', debug_info)
        self.assertIn('methods', debug_info)
        self.assertGreater(len(debug_info['methods']), 0)

        for debug_method in debug_info['methods']:
            self.assertIn('name', debug_method)
            parsed_name = debug_method['name'].split(',')
            self.assertEqual(2, len(parsed_name))
            self.assertIn(parsed_name[-1], methods)
            actual_method = methods[parsed_name[-1]]

            # validate id
            self.assertIn('id', debug_method)
            self.assertEqual(str(id(actual_method)), debug_method['id'])

            # validate parameters
            self.assertIn('params', debug_method)
            self.assertEqual(len(actual_method.args), len(debug_method['params']))
            for var in debug_method['params']:
                self.assertEqual(2, len(var.split(',')))
                param_id, param_type = var.split(',')
                self.assertIn(param_id, actual_method.args)
                self.assertEqual(param_type, actual_method.args[param_id].type.abi_type)

            # validate local variables
            self.assertIn('variables', debug_method)
            self.assertEqual(len(actual_method.locals), len(debug_method['variables']))
            for var in debug_method['variables']:
                self.assertEqual(2, len(var.split(',')))
                var_id, var_type = var.split(',')
                self.assertIn(var_id, actual_method.locals)
                self.assertEqual(actual_method.locals[var_id].type.abi_type, var_type)

    def test_generate_nefdbgnfo_file_with_event(self):
        path = '%s/boa3_test/test_sc/event_test/EventWithArgument.py' % self.dirname

        expected_nef_output = path.replace('.py', '.nefdbgnfo')
        compiler = Compiler()
        compiler.compile_and_save(path, path.replace('.py', '.nef'))
        events: Dict[str, Event] = {
            name: method
            for name, method in self.get_compiler_analyser(compiler).symbol_table.items()
            if isinstance(method, Event)
        }

        self.assertTrue(os.path.exists(expected_nef_output))
        debug_info = self.get_debug_info(path)
        self.assertNotIn('entrypoint', debug_info)
        self.assertIn('events', debug_info)
        self.assertGreater(len(debug_info['events']), 0)

        for debug_event in debug_info['events']:
            self.assertIn('name', debug_event)
            parsed_name = debug_event['name'].split(',')
            self.assertEqual(2, len(parsed_name))
            self.assertIn(parsed_name[-1], events)
            actual_event = events[parsed_name[-1]]

            # validate id
            self.assertIn('id', debug_event)
            self.assertEqual(str(id(actual_event)), debug_event['id'])

            # validate parameters
            self.assertIn('params', debug_event)
            self.assertEqual(len(actual_event.args), len(debug_event['params']))
            for var in debug_event['params']:
                self.assertEqual(2, len(var.split(',')))
                param_id, param_type = var.split(',')
                self.assertIn(param_id, actual_event.args)
                self.assertEqual(param_type, actual_event.args[param_id].type.abi_type)

    def test_generate_manifest_file_with_notify_event(self):
        path = '%s/boa3_test/test_sc/interop_test/NotifySequence.py' % self.dirname
        expected_manifest_output = path.replace('.py', '.manifest.json')
        output, manifest = self.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('events', abi)
        self.assertGreater(len(abi['events']), 0)

        notify_event = next(abi_event for abi_event in abi['events']
                            if 'name' in abi_event and abi_event['name'] == 'notify')
        self.assertIsNotNone(notify_event, "notify event is not listed in the contract's abi")
        self.assertIn('parameters', notify_event)
        self.assertEqual(1, len(notify_event['parameters']))
        self.assertIn('type', notify_event['parameters'][0])
        self.assertEqual(AbiType.Any, notify_event['parameters'][0]['type'])

    def test_generate_without_main(self):
        path = '%s/boa3_test/test_sc/generation_test/GenerationWithoutMain.py' % self.dirname
        expected_manifest_output = path.replace('.py', '.manifest.json')
        output, manifest = self.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertNotIn('entryPoint', abi)
        self.assertIn('methods', abi)
        self.assertEqual(2, len(abi['methods']))

        self.assertIn('events', abi)
        self.assertEqual(0, len(abi['events']))

    def test_generate_without_main_and_public_methods(self):
        path = '%s/boa3_test/test_sc/generation_test/GenerationWithoutMainAndPublicMethods.py' % self.dirname
        expected_manifest_output = path.replace('.py', '.manifest.json')
        output, manifest = self.compile_and_save(path)

        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertNotIn('entryPoint', abi)
        self.assertIn('methods', abi)
        self.assertEqual(0, len(abi['methods']))

        self.assertIn('events', abi)
        self.assertEqual(0, len(abi['events']))

    def test_generate_manifest_file_abi_method_offset(self):
        path = '%s/boa3_test/test_sc/generation_test/GenerationWithDecorator.py' % self.dirname
        manifest_path = path.replace('.py', '.manifest.json')

        compiler = Compiler()
        compiler.compile_and_save(path, path.replace('.py', '.nef'))
        methods: Dict[str, Method] = {
            name: method
            for name, method in self.get_compiler_analyser(compiler).symbol_table.items()
            if isinstance(method, Method)
        }
        self.assertGreater(len(methods), 0)

        output, manifest = self.get_output(path)
        self.assertTrue(os.path.exists(manifest_path))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('methods', abi)
        abi_methods = abi['methods']
        self.assertGreater(len(abi['methods']), 0)

        for method in abi_methods:
            self.assertIn('name', method)
            self.assertIn('offset', method)
            self.assertIn(method['name'], methods)
            self.assertEqual(method['offset'], methods[method['name']].start_address)

    def test_generate_manifest_file_storage_feature(self):
        path = '%s/boa3_test/test_sc/storage_test/StorageGetBytesKey.py' % self.dirname
        manifest_path = path.replace('.py', '.manifest.json')

        output, manifest = self.compile_and_save(path)
        self.assertTrue(os.path.exists(manifest_path))

        self.assertIn('features', manifest)
        self.assertIn('storage', manifest['features'])
        self.assertEqual(True, manifest['features']['storage'])

    def test_generate_debug_info_with_multiple_flows(self):
        path = '%s/boa3_test/test_sc/generation_test/GenerationWithMultipleFlows.py' % self.dirname

        compiler = Compiler()
        compiler.compile_and_save(path, path.replace('.py', '.nef'))
        methods: Dict[str, Method] = {
            name: method
            for name, method in self.get_compiler_analyser(compiler).symbol_table.items()
            if isinstance(method, Method)
        }
        self.assertGreater(len(methods), 0)

        debug_info = self.get_debug_info(path)
        self.assertIn('methods', debug_info)
        self.assertGreater(len(debug_info['methods']), 0)

        for debug_method in debug_info['methods']:
            self.assertIn('name', debug_method)
            parsed_name = debug_method['name'].split(',')
            self.assertEqual(2, len(parsed_name))
            self.assertIn(parsed_name[-1], methods)
            actual_method = methods[parsed_name[-1]]

            # validate id
            self.assertIn('id', debug_method)
            self.assertEqual(str(id(actual_method)), debug_method['id'])

            # validate parameters
            self.assertIn('params', debug_method)
            self.assertEqual(len(actual_method.args), len(debug_method['params']))
            for var in debug_method['params']:
                self.assertEqual(2, len(var.split(',')))
                param_id, param_type = var.split(',')
                self.assertIn(param_id, actual_method.args)
                self.assertEqual(param_type, actual_method.args[param_id].type.abi_type)

            # validate local variables
            self.assertIn('variables', debug_method)
            self.assertEqual(len(actual_method.locals), len(debug_method['variables']))
            for var in debug_method['variables']:
                self.assertEqual(2, len(var.split(',')))
                var_id, var_type = var.split(',')
                self.assertIn(var_id, actual_method.locals)
                self.assertEqual(actual_method.locals[var_id].type.abi_type, var_type)

    def test_generate_init_method(self):
        path = '%s/boa3_test/test_sc/variable_test/GlobalAssignmentWithType.py' % self.dirname

        compiler = Compiler()
        compiler.compile_and_save(path, path.replace('.py', '.nef'))
        methods: Dict[str, Method] = {
            name: method
            for name, method in self.get_compiler_analyser(compiler).symbol_table.items()
            if isinstance(method, Method)
        }
        from boa3.constants import INITIALIZE_METHOD_ID
        self.assertGreater(len(methods), 0)
        self.assertIn(INITIALIZE_METHOD_ID, methods)
        init_method = methods[INITIALIZE_METHOD_ID]

        output, manifest = self.get_output(path)
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('methods', abi)
        self.assertGreater(len(abi['methods']), 0)

        abi_init = next(method for method in abi['methods']
                        if 'name' in method and method['name'] == INITIALIZE_METHOD_ID)
        self.assertIsNotNone(abi_init)
        self.assertIn('offset', abi_init)
        self.assertEqual(init_method.start_address, abi_init['offset'])
        self.assertIn('parameters', abi_init)
        self.assertEqual(0, len(abi_init['parameters']))
        self.assertIn('returntype', abi_init)
        self.assertEqual(AbiType.Void, abi_init['returntype'])

        debug_info = self.get_debug_info(path)
        self.assertIn('methods', debug_info)
        self.assertGreater(len(debug_info['methods']), 0)

        debug_method = next((method for method in debug_info['methods']
                             if 'id' in method and method['id'] == str(id(init_method))), None)
        self.assertIsNotNone(debug_method)
        parsed_name = debug_method['name'].split(',')
        self.assertEqual(2, len(parsed_name))
        self.assertIn(parsed_name[-1], methods)

        # validate parameters
        self.assertIn('params', debug_method)
        self.assertEqual(0, len(debug_method['params']))

        # validate local variables
        self.assertIn('variables', debug_method)
        self.assertEqual(0, len(debug_method['variables']))

        # validate sequence points
        self.assertIn('sequence-points', debug_method)
        self.assertEqual(0, len(debug_method['sequence-points']))



import ast
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, List

from boa3.model.builtin.builtincallable import IBuiltinCallable
from boa3.model.method import Method
from boa3.model.type.itype import IType
from boa3.model.variable import Variable


class IBuiltinMethod(IBuiltinCallable, Method, ABC):
    def __init__(self, identifier: str, args: Dict[str, Variable] = None,
                 defaults: List[ast.AST] = None, return_type: IType = None):
        super().__init__(identifier, args, defaults, return_type)

    @property
    def is_supported(self) -> bool:
        """
        Verifies if the builtin method is supported by the compiler

        :return: True if it is supported. False otherwise.
        """
        return True

    @property
    def args_on_stack(self) -> int:
        """
        Gets the number of arguments that must be on stack before the opcode is called.

        :return: the number of arguments if opcode is not empty. Zero otherwise.
        """
        if len(self.opcode) > 0:
            num_args = self._args_on_stack
            if num_args < 0:
                return 0
            elif num_args > len(self.args):
                return len(self.args)
            return num_args
        else:
            return 0

    def push_self_first(self) -> bool:
        """
        Verifies if the `self` value of the method needs to be pushed to the Neo execution stack before the
        other arguments.

        :return: a boolean value indicating if the `self` argument must be pushed before. Returns False if there isn't
                 a `self` argument in the function
        """
        return False

    def validate_self(self, self_type: IType) -> bool:
        """
        Verifies if the given value is valid to the function `self` argument

        :param self_type: type of the value
        :return: a boolean value that represents if the value is valid. Returns False if there isn't a `self` argument
                 in the function
        """
        if not self.has_self_argument:
            return False
        return self.args['self'].type.is_type_of(self_type)

    @property
    def has_self_argument(self) -> bool:
        """
        Verifies if the function has a `self` argument.

        :return: True if there is this argument. False otherwise.
        """
        return 'self' in self.args

    @property
    @abstractmethod
    def _args_on_stack(self) -> int:
        """
        Gets the number of arguments that must be on stack before the opcode is called.

        :return: the number of arguments.
        """
        return 0

    @property
    def stores_on_slot(self) -> bool:
        """
        Returns whether this method needs to update the value from a variable

        :return: the number of arguments.
        """
        return False

    @property
    def body(self) -> Optional[str]:
        """
        Gets the body of the method.

        :return: Return the code of the method body if there is no opcode. None otherwise.
        """
        return self._body if len(self.opcode) <= 0 else None

    @property
    @abstractmethod
    def _body(self) -> Optional[str]:
        """
        Gets the body of the method.

        :return: Return the code of the method body.
        """
        return None

    def build(self, value: Any):
        """
        Creates a method instance with the given value as self

        :param value: value to build the type
        :return: The built method if the value is valid. The current object otherwise
        :rtype: IBuiltinMethod
        """
        return self



import ast
import logging
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from boa3.analyser.astanalyser import IAstAnalyser
from boa3.analyser.importanalyser import ImportAnalyser
from boa3.builtin import NeoMetadata
from boa3.exception import CompilerError, CompilerWarning
from boa3.model.builtin.builtin import Builtin
from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.callable import Callable
from boa3.model.event import Event
from boa3.model.expression import IExpression
from boa3.model.importsymbol import Import
from boa3.model.method import Method
from boa3.model.module import Module
from boa3.model.symbol import ISymbol
from boa3.model.type.collection.icollection import ICollectionType as Collection
from boa3.model.type.collection.sequence.sequencetype import SequenceType
from boa3.model.type.type import IType, Type
from boa3.model.variable import Variable


class ModuleAnalyser(IAstAnalyser, ast.NodeVisitor):
    """
    This class is responsible for mapping the locals of the functions and modules

    The methods with the name starting with 'visit_' are implementations of methods from the :class:`NodeVisitor` class.
    These methods are used to walk through the Python abstract syntax tree.

    :ivar modules: a dictionary that maps each module with its name. Empty by default.
    :ivar symbols: a dictionary that maps the global symbols.
    """

    def __init__(self, analyser, symbol_table: Dict[str, ISymbol], filename: str = None, log: bool = False):
        super().__init__(analyser.ast_tree, filename, log)
        self.modules: Dict[str, Module] = {}
        self.symbols: Dict[str, ISymbol] = symbol_table

        self._builtin_functions_to_visit: Dict[str, IBuiltinMethod] = {}
        self._current_module: Module = None
        self._current_method: Method = None
        self._current_event: Event = None

        self._metadata: NeoMetadata = None
        self._metadata_node: ast.AST = ast.parse('')
        self.imported_nodes: List[ast.AST] = []
        self.visit(self._tree)

        self._validate_metadata_symbols()
        analyser.metadata = self._metadata if self._metadata is not None else NeoMetadata()

    @property
    def _current_scope(self):
        """
        Returns the scope that is currently being analysed

        :return: the current scope. Return None if it is the global scope
        :rtype: Method or Module or None
        """
        if self._current_method is not None:
            return self._current_method
        return self._current_module

    @property
    def global_symbols(self) -> Dict[str, ISymbol]:
        """
        Returns the symbols in global scope

        :return: the symbol table of the global scope
        """
        global_symbols: Dict[str, ISymbol] = {}

        global_symbols.update(self.symbols)
        for mod in self.modules.values():
            global_symbols.update(mod.symbols)

        return global_symbols

    def __include_variable(self, var_id: str, var_type_id: Union[str, IType],
                           source_node: ast.AST,
                           var_enumerate_type: IType = Type.none, assignment: bool = True):
        """
        Includes the variable in the symbol table if the id was not used

        :param var_id: variable id
        :param var_type_id: variable type id
        :type var_type_id: str or IType
        :param var_enumerate_type: variable value type id if var_type_id is a SequenceType
        """
        if not isinstance(var_id, str) and isinstance(var_id, Iterable):
            variables = [var_name.id for var_name in source_node.targets[0].elts]
            var_types = self.visit(source_node.value)
        else:
            variables = [var_id]
            var_types = [var_type_id]

        for x in range(min(len(variables), len(var_types))):
            var_id, var_type_id = variables[x], var_types[x]
            if var_id not in self._current_scope.symbols:
                outer_symbol = self.get_symbol(var_id)
                if outer_symbol is not None:
                    self._log_warning(
                        CompilerWarning.NameShadowing(source_node.lineno, source_node.col_offset, outer_symbol, var_id)
                    )

                var_type = None
                if isinstance(var_type_id, SequenceType):
                    var_type = var_type_id
                    var_enumerate_type = self.get_enumerate_type(var_type)
                elif isinstance(var_type_id, IType):
                    var_type = var_type_id

                # when setting a None value to a variable, set the variable as any type
                if var_type is Type.none:
                    var_type = Type.any

                if isinstance(var_type, IType) or var_type is None:
                    # if type is None, the variable type depends on the type of a expression
                    if isinstance(var_type, SequenceType):
                        var_type = var_type.build_collection(var_enumerate_type)
                    var = Variable(var_type, origin_node=source_node)
                    self._current_scope.include_variable(var_id, var)
            if hasattr(self._current_scope, 'assign_variable') and assignment:
                self._current_scope.assign_variable(var_id)

    def __include_callable(self, callable_id: str, callable: Callable):
        """
        Includes the method in the symbol table if the id was not used

        :param callable_id: method id
        :param callable: method to be included
        """
        if callable_id not in self._current_module.symbols:
            self._current_module.include_callable(callable_id, callable)

    def get_symbol(self, symbol_id: str) -> Optional[ISymbol]:
        if symbol_id in self._current_scope.symbols:
            # the symbol exists in the local scope
            return self._current_scope.symbols[symbol_id]
        elif symbol_id in self._current_module.symbols:
            # the symbol exists in the module scope
            return self._current_module.symbols[symbol_id]
        else:
            return super().get_symbol(symbol_id)

    # region Log

    def _log_import(self, import_from: str):
        if self._log:
            logging.info("Importing '{0}'".format(import_from))

    def _log_unresolved_import(self, origin_node: ast.AST, import_id: str):
        if self._log:
            self._log_error(
                CompilerError.UnresolvedReference(
                    line=origin_node.lineno,
                    col=origin_node.col_offset,
                    symbol_id=import_id
                )
            )

    # endregion

    # region Metadata

    def _read_metadata_object(self, function: ast.FunctionDef):
        """
        Gets the metadata object defined in this function

        :param function:
        """
        if self._metadata is not None:
            # metadata function has been defined already
            self._log_warning(
                CompilerWarning.RedeclaredSymbol(
                    line=function.lineno, col=function.col_offset,
                    symbol_id=Builtin.Metadata.identifier
                )
            )
        # this function must have a return and no arguments
        elif len(function.args.args) != 0:
            self._log_error(
                CompilerError.UnexpectedArgument(
                    line=function.lineno, col=function.col_offset
                )
            )
        elif not any(isinstance(stmt, ast.Return) for stmt in function.body):
            self._log_error(
                CompilerError.MissingReturnStatement(
                    line=function.lineno, col=function.col_offset,
                    symbol_id=function.name
                )
            )
        else:
            function.returns = None
            function.decorator_list = []
            module: ast.Module = ast.parse('')
            module.body = [node for node in self._tree.body
                           if isinstance(node, (ast.ImportFrom, ast.Import))]
            module.body.append(function)
            ast.copy_location(module, function)

            # executes the function
            code = compile(module, filename='<boa3>', mode='exec')
            namespace = {}
            exec(code, namespace)
            obj: Any = namespace[function.name]()

            node: ast.AST = function.body[-1] if len(function.body) > 0 else function
            # return must be a NeoMetadata object
            if not isinstance(obj, NeoMetadata):
                obj_type = self.get_type(obj).identifier if self.get_type(obj) is not Type.any else type(obj).__name__
                self._log_error(
                    CompilerError.MismatchedTypes(
                        line=node.lineno, col=node.col_offset,
                        expected_type_id=NeoMetadata.__name__,
                        actual_type_id=obj_type
                    )
                )
                return

            # validates the metadata attributes types
            attributes: Dict[str, Any] = {attr: value
                                          for attr, value in dict(obj.__dict__).items()
                                          if attr in Builtin.metadata_fields}
            if any(not isinstance(value, Builtin.metadata_fields[attr]) for attr, value in attributes.items()):
                for expected, actual in [(Builtin.metadata_fields[attr], type(v_type))
                                         for attr, v_type in attributes.items()
                                         if not isinstance(v_type, Builtin.metadata_fields[attr])]:
                    if isinstance(expected, Iterable):
                        expected_id = 'Union[{0}]'.format(', '.join([tpe.__name__ for tpe in expected]))
                    elif hasattr(expected, '__name__'):
                        expected_id = expected.__name__
                    else:
                        expected_id = str(expected)

                    self._log_error(
                        CompilerError.MismatchedTypes(
                            line=node.lineno, col=node.col_offset,
                            expected_type_id=expected_id,
                            actual_type_id=actual.__name__
                        )
                    )
            else:
                # if the function was defined correctly, sets the metadata object of the smart contract
                self._metadata = obj
                self._metadata_node = function  # for error messages only

    def _validate_metadata_symbols(self):
        """
        Validates if the symbols match the requirements from the given metadata
        """
        if self._metadata is not None:
            if self._metadata.is_payable:
                """
                Payable smart contracts requires the implementation of a 'verify' function, that must implement the
                following signature:

                ```
                @public
                def verify(*args) -> bool
                ```
                """
                verify_id: str = 'verify'
                if (verify_id not in self.global_symbols  # couldn't find verify
                        or not isinstance(self.global_symbols[verify_id], Method)  # verify is not a function
                        or self.global_symbols[verify_id].origin is None  # verify is not a user function
                    ):
                    self._log_error(
                        CompilerError.MetadataImplementationMissing(
                            line=self._metadata_node.lineno, col=self._metadata_node.col_offset,
                            symbol_id=verify_id,
                            metadata_attr_id='is_payable'
                        )
                    )
                elif (self.global_symbols[verify_id].return_type != Type.bool
                      or not self.global_symbols[verify_id].is_public
                      ):
                    actual = self.global_symbols[verify_id]
                    expected = Method(actual.args, Type.bool, True)
                    self._log_error(
                        CompilerError.MetadataIncorrectImplementation(
                            line=actual.origin.lineno, col=actual.origin.col_offset,
                            symbol_id=verify_id,
                            expected_symbol=expected,
                            actual_symbol=actual
                        )
                    )

    # endregion

    # region AST

    def visit_ImportFrom(self, import_from: ast.ImportFrom):
        """
        Includes methods and variables from other modules into the current scope

        :param import_from:
        """
        self._log_import(import_from.module)
        analyser = self._analyse_module_to_import(import_from, import_from.module)
        if analyser is not None:
            import_alias: Dict[str] = \
                {alias.name: alias.asname if alias.asname is not None else alias.name for alias in import_from.names}

            new_symbols: Dict[str, ISymbol] = analyser.export_symbols(list(import_alias.keys()))
            # includes the module to be able to generate the functions
            imported_module = Import(analyser.path, analyser.tree, analyser, import_alias)
            self._current_scope.include_symbol(import_from.module, imported_module)

            for name, alias in import_alias.items():
                if name in new_symbols:
                    self._current_scope.include_symbol(alias, imported_module.symbols[name])
                else:
                    # if there's a symbol that couldn't be loaded, log a compiler error
                    self._log_unresolved_import(import_from, name)

    def _analyse_module_to_import(self, origin_node: ast.AST, target: str) -> Optional[ImportAnalyser]:
        analyser = ImportAnalyser(target)
        if analyser.can_be_imported:
            return analyser
        else:
            self._log_unresolved_import(origin_node, target)

    def visit_Import(self, import_node: ast.Import):
        """
        Includes methods and variables from other modules into the current scope

        :param import_node:
        """
        import_alias: Dict[str] = \
            {alias.name: alias.asname if alias.asname is not None else alias.name for alias in import_node.names}

        for target, alias in import_alias.items():
            self._log_import(target)
            analyser = self._analyse_module_to_import(import_node, target)
            if analyser is not None:
                new_symbols: Dict[str, ISymbol] = analyser.export_symbols()
                for symbol in [symbol for symbol in analyser.symbols if symbol not in new_symbols]:
                    # if there's a symbol that couldn't be loaded, log a compiler error
                    self._log_unresolved_import(import_node, '{0}.{1}'.format(target, symbol))

                imported_module = Import(analyser.path, analyser.tree, analyser)
                self._current_scope.include_symbol(alias, imported_module)

    def visit_Module(self, module: ast.Module):
        """
        Visitor of the module node

        Fills module symbol table

        :param module:
        """
        mod: Module = Module()
        self._current_module = mod

        global_stmts = []
        function_stmts = []
        for stmt in module.body:
            if isinstance(stmt, (ast.FunctionDef, ast.AsyncFunctionDef)):
                function_stmts.append(stmt)
            elif not (isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.Constant)):
                # don't evaluate constant expression - for example: string for documentation
                if self.visit(stmt) is not Builtin.Event:
                    global_stmts.append(stmt)

        module.body = global_stmts + function_stmts
        for var_id, var in mod.variables.items():
            # all static fields must be initialized
            if not mod.is_variable_assigned(var_id):
                self._log_error(
                    CompilerError.UnresolvedReference(
                        line=var.origin.lineno if var.origin is not None else 0,
                        col=var.origin.col_offset if var.origin is not None else 0,
                        symbol_id=var_id
                    )
                )

        for stmt in function_stmts:
            result = self.visit(stmt)
            # don't evaluate the metadata function in the following analysers
            if result is Builtin.Metadata:
                module.body.remove(stmt)

        # TODO: include the body of the builtin methods to the ast
        # TODO: get module name
        self.modules['main'] = mod
        self._current_module = None

    def visit_FunctionDef(self, function: ast.FunctionDef):
        """
        Visitor of the function node

        Includes the method in the scope of its module

        :param function:
        """
        fun_args = self.visit(function.args)
        fun_rtype_symbol = self.visit(function.returns) if function.returns is not None else Type.none

        if fun_rtype_symbol is None:
            # it is a function with None return: Main(a: int) -> None:
            raise NotImplementedError

        if isinstance(fun_rtype_symbol, str):
            symbol = self.get_symbol(function.returns.id)
            fun_rtype_symbol = self.get_type(symbol)

        fun_return: IType = self.get_type(fun_rtype_symbol)
        fun_decorators: List[Method] = self._get_function_decorators(function)

        if Builtin.Metadata in fun_decorators:
            self._read_metadata_object(function)
            return Builtin.Metadata

        method = Method(args=fun_args, defaults=function.args.defaults, return_type=fun_return,
                        origin_node=function, is_public=Builtin.Public in fun_decorators)
        self._current_method = method

        # don't evaluate constant expression - for example: string for documentation
        function.body = [stmt for stmt in function.body
                         if not (isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.Constant))]
        for stmt in function.body:
            self.visit(stmt)

        self.__include_callable(function.name, method)
        self._current_method = None

    def _get_function_decorators(self, function: ast.FunctionDef) -> List[Method]:
        """
        Gets a list of the symbols used to decorate the given function

        :param function: python ast function definition node
        :return: a list with all function decorators. Empty if none decorator is found.
        """
        return [self.get_symbol(self.visit(decorator)) for decorator in function.decorator_list]

    def visit_arguments(self, arguments: ast.arguments) -> Dict[str, Variable]:
        """
        Visitor of the function arguments node

        :param arguments:
        :return: a dictionary that maps each argument to its identifier
        """
        args: Dict[str, Variable] = {}

        for arg in arguments.args:
            var_id, var = self.visit_arg(arg)  # Tuple[str, Variable]
            args[var_id] = var
        return args

    def visit_arg(self, arg: ast.arg) -> Tuple[str, Variable]:
        """
        Visitor of a function argument node

        :param arg:
        :return: a tuple with the identifier and the argument
        """
        var_id = arg.arg
        var_type: IType = self.get_type(arg.annotation)

        if var_type is Type.none and isinstance(arg.annotation, ast.Name):
            var_symbol: ISymbol = self.get_symbol(arg.annotation.id)
            var_type = self.get_type(var_symbol)

        return var_id, Variable(var_type)

    def visit_Return(self, ret: ast.Return):
        """
        Visitor of the function return node

        If the return is a name, verifies if the symbol is defined

        :param ret: the python ast return node
        """
        if isinstance(ret.value, ast.Name):
            symbol_id = self.visit(ret.value)
            symbol = self.get_symbol(symbol_id)
            if symbol is None:
                # the symbol doesn't exists
                self._log_error(
                    CompilerError.UnresolvedReference(ret.value.lineno, ret.value.col_offset, symbol_id)
                )

    def visit_type(self, target: ast.AST) -> Optional[IType]:
        """
        Gets the type by its identifier

        :param target: ast node to be evaluated
        :return: the type of the value inside the node. None by default
        """
        target_type = self.visit(target)  # Type:str or IType
        if isinstance(target_type, str) and not isinstance(target, ast.Str):
            symbol = self.get_symbol(target_type)
            if symbol is None:
                # the symbol doesn't exists
                self._log_error(
                    CompilerError.UnresolvedReference(target.lineno, target.col_offset, target_type)
                )
            target_type = symbol

        if target_type is None and not isinstance(target, ast.NameConstant):
            # the value type is invalid
            return None

        if not isinstance(target_type, (ast.AST, IType, IExpression)) and isinstance(target_type, ISymbol):
            self._log_error(
                CompilerError.UnresolvedReference(
                    line=target.lineno, col=target.col_offset,
                    symbol_id=str(target_type)
                )
            )
        return self.get_type(target_type)

    def get_enumerate_type(self, var_type: IType) -> IType:
        return var_type.value_type if isinstance(var_type, SequenceType) else Type.none

    def visit_Assign(self, assign: ast.Assign):
        """
        Visitor of the variable assignment node

        Includes the variable in its scope if it's the first use

        :param assign:
        """
        var_id = self.visit(assign.targets[0])
        var_type = self.visit_type(assign.value)

        if var_type is Type.none and isinstance(assign.value, ast.Name):
            symbol = self.get_symbol(assign.value.id)
            if isinstance(symbol, Event):
                var_type = Builtin.Event
                self._current_event = symbol

        return self.assign_value(var_id, var_type, source_node=assign)

    def visit_AnnAssign(self, ann_assign: ast.AnnAssign):
        """
        Visitor of the annotated variable assignment node

        Includes the variable in its scope if it's the first use

        :param ann_assign:
        """
        var_id: str = self.visit(ann_assign.target)
        var_type: IType = self.visit_type(ann_assign.annotation)

        # TODO: check if the annotated type and the value type are the same
        return self.assign_value(var_id, var_type, source_node=ann_assign, assignment=ann_assign.value is not None)

    def assign_value(self, var_id: str, var_type: IType, source_node: ast.AST, assignment: bool = True) -> IType:
        if var_type is Builtin.Event and self._current_event is not None:
            if '' in self._current_module.symbols and self._current_module.symbols[''] is self._current_event:
                self._current_scope.callables[var_id] = self._current_scope.callables.pop('')
                self._current_event.name = var_id
            else:
                self._current_scope.callables[var_id] = self._current_event
            self._current_event = None
        else:
            self.__include_variable(var_id, var_type, source_node=source_node, assignment=assignment)
        return var_type

    def visit_Global(self, global_node: ast.Global):
        """
        Visitor of the global identifier node

        :param global_node:
        """
        raise NotImplementedError

    def visit_Expr(self, expr: ast.Expr):
        """
        Visitor of the atom expression node

        :param expr:
        """
        value = self.visit(expr.value)
        if isinstance(expr.value, ast.Name):
            if (
                    # it is not a symbol of the method scope
                    (self._current_method is not None and value not in self._current_method.symbols)
                    # nor it is a symbol of the module scope
                    or (self._current_module is not None and value not in self._current_module.symbols)
                    # nor it is a symbol of the global scope
                    or value not in self.symbols
            ):
                self._log_error(
                    CompilerError.UnresolvedReference(expr.value.lineno, expr.value.col_offset, value)
                )

    def visit_Subscript(self, subscript: ast.Subscript):
        """
        Verifies if it is the types in the subscription are valid

        :param subscript: the python ast subscription node
        :return: if the subscript is not a symbol, returns its type. Otherwise returns the symbol id.
        :rtype: IType or str
        """
        value = self.visit(subscript.value)
        symbol = self.get_symbol(value) if isinstance(value, str) else value

        if isinstance(subscript.ctx, ast.Load):
            if isinstance(symbol, Collection):
                values_type: Iterable[IType] = self.get_values_type(subscript.slice.value)
                return symbol.build_collection(*values_type)

            symbol_type = self.get_type(symbol)
            if isinstance(symbol_type, SequenceType):
                return symbol_type.value_type

        return value

    def get_values_type(self, value: ast.AST) -> Iterable[Optional[IType]]:
        """
        Verifies if it is a multiple assignments statement

        :param value: the python ast subscription node
        """
        value_type: Optional[IType] = None

        if isinstance(value, (ast.Subscript, ast.Attribute, ast.Tuple)):
            # index is another subscription
            value_type = self.visit(value)
        elif isinstance(value, ast.Name):
            # index is an identifier
            value_type = self.get_symbol(value.id)

        types: Iterable[Optional[IType]] = value_type if isinstance(value_type, Iterable) else [value_type]
        for tpe in types:
            if not isinstance(tpe, IType):
                # type hint not using identifiers or using identifiers that are not types
                index = self.visit(value)
                self._log_error(
                    CompilerError.UnresolvedReference(value.lineno, value.col_offset, index)
                )

        return types

    def visit_Call(self, call: ast.Call) -> Optional[IType]:
        """
        Visitor of a function call node

        :param call: the python ast function call node
        :return: the result type of the called function. None if the function is not found
        """
        func_id = self.visit(call.func)
        func_symbol = self.get_symbol(func_id)

        # if func_symbol is None, the called function may be a function written after in the code
        # that's why it shouldn't log a compiler error here
        if func_symbol is None:
            return None

        if not isinstance(func_symbol, Callable):
            # verifiy if it is a builtin method with its name shadowed
            func = Builtin.get_symbol(func_id)
            func_symbol = func if func is not None else func_symbol
        if not isinstance(func_symbol, Callable):
            # the symbol doesn't exists
            self._log_error(
                CompilerError.UnresolvedReference(call.func.lineno, call.func.col_offset, func_id)
            )
        elif isinstance(func_symbol, IBuiltinMethod):
            if func_symbol.body is not None:
                self._builtin_functions_to_visit[func_id] = func_symbol
            elif func_symbol is Builtin.NewEvent:
                new_event = self.create_new_event(call)
                self.__include_callable(new_event.identifier, new_event)
                self._current_event = new_event

        return self.get_type(call.func)

    def create_new_event(self, create_call: ast.Call) -> Event:
        event = Event('')
        event_args = create_call.args

        if len(event_args) < 0:
            self._log_error(
                CompilerError.UnfilledArgument(line=create_call.lineno,
                                               col=create_call.col_offset,
                                               param=list(Builtin.NewEvent.args)[0])
            )
        elif len(event_args) > 0:
            args_type = self.get_type(event_args[0])
            if not Type.list.is_type_of(args_type):
                self._log_error(
                    CompilerError.MismatchedTypes(line=event_args[0].lineno,
                                                  col=event_args[0].col_offset,
                                                  expected_type_id=Type.list.identifier,
                                                  actual_type_id=args_type.identifier)
                )
            else:
                for value in event_args[0].elts:
                    if not isinstance(value, ast.Tuple):
                        CompilerError.MismatchedTypes(line=value.lineno,
                                                      col=value.col_offset,
                                                      expected_type_id=Type.tuple.identifier,
                                                      actual_type_id=self.get_type(value).identifier)
                    elif len(value.elts) < 2:
                        self._log_error(
                            CompilerError.UnfilledArgument(line=value.lineno,
                                                           col=value.col_offset,
                                                           param=list(Builtin.NewEvent.args)[0])
                        )
                    elif not (isinstance(value.elts[0], ast.Str)
                              and (isinstance(value.elts[1], ast.Name)
                                   and isinstance(self.get_symbol(value.elts[1].id), IType)
                                   )):
                        CompilerError.MismatchedTypes(line=value.lineno,
                                                      col=value.col_offset,
                                                      expected_type_id=Type.tuple.identifier,
                                                      actual_type_id=self.get_type(value).identifier)
                    else:
                        arg_name = value.elts[0].s
                        arg_type = self.get_symbol(value.elts[1].id)
                        event.args[arg_name] = Variable(arg_type)

            if len(event_args) > 1:
                if not isinstance(event_args[1], ast.Str):
                    name_type = self.get_type(event_args[1])
                    CompilerError.MismatchedTypes(line=event_args[1].lineno,
                                                  col=event_args[1].col_offset,
                                                  expected_type_id=Type.str.identifier,
                                                  actual_type_id=name_type.identifier)
                else:
                    event.name = event_args[1].s

        return event

    def visit_Attribute(self, attribute: ast.Attribute) -> Union[ISymbol, str]:
        """
        Gets the attribute inside the ast node

        :param attribute: the python ast attribute node
        :return: returns the type of the value, the attribute symbol and its id if the attribute exists.
                 Otherwise, returns None
        """
        value_id = attribute.value.id if isinstance(attribute.value, ast.Name) else None
        value: ISymbol = self.get_symbol(value_id) if value_id is not None else self.visit(attribute.value)

        if hasattr(value, 'symbols') and attribute.attr in value.symbols:
            return value.symbols[attribute.attr]
        elif Builtin.get_symbol(attribute.attr) is not None:
            return Builtin.get_symbol(attribute.attr)
        else:
            return '{0}.{1}'.format(value_id, attribute.attr)

    def visit_For(self, for_node: ast.For):
        """
        Visitor of for statement node

        :param for_node: the python ast for node
        """
        iter_type = self.get_type(for_node.iter)
        targets = self.visit(for_node.target)

        if isinstance(iter_type, SequenceType):
            if isinstance(targets, str):
                self.__include_variable(targets, iter_type.value_type, source_node=for_node.target)
            else:
                for target in targets:
                    if isinstance(target, str):
                        self.__include_variable(target, iter_type.value_type, source_node=for_node.target)

        # continue to walk through the tree
        self.generic_visit(for_node)

    def visit_Name(self, name: ast.Name) -> str:
        """
        Visitor of a name node

        :param name:
        :return: the identifier of the name
        """
        return name.id

    def visit_Constant(self, constant: ast.Constant) -> Any:
        """
        Visitor of constant values node

        :param constant: the python ast constant value node
        :return: the value of the constant
        """
        return constant.value

    def visit_NameConstant(self, constant: ast.NameConstant) -> Any:
        """
        Visitor of constant names node

        :param constant:
        :return: the value of the constant
        """
        return constant.value

    def visit_Num(self, num: ast.Num) -> int:
        """
        Visitor of literal number node

        :param num:
        :return: the value of the number
        """
        return num.n

    def visit_Str(self, str: ast.Str) -> str:
        """
        Visitor of literal string node

        :param str:
        :return: the value of the string
        """
        return str.s

    def visit_Tuple(self, tup_node: ast.Tuple) -> Tuple[Any, ...]:
        """
        Visitor of literal tuple node

        :param tup_node: the python ast string node
        :return: the value of the tuple
        """
        return tuple([self.get_type(value) for value in tup_node.elts])

    def visit_List(self, list_node: ast.List) -> List[Any]:
        """
        Visitor of literal list node

        :param list_node: the python ast list node
        :return: the value of the list
        """
        return [self.get_type(value) for value in list_node.elts]

    def visit_Dict(self, dict_node: ast.Dict) -> Dict[Any, Any]:
        """
        Visitor of literal dict node

        :param dict_node: the python ast dict node
        :return: the value of the dict
        """
        dictionary = {}
        size = min(len(dict_node.keys), len(dict_node.values))
        for index in range(size):
            key = self.get_type(dict_node.keys[index])
            value = self.get_type(dict_node.values[index])
            if key in dictionary and dictionary[key] != value:
                dictionary[key] = Type.get_generic_type(dictionary[key], value)
            else:
                dictionary[key] = value
        return dictionary

    # endregion



from __future__ import annotations

import asyncio
import ipaddress
from contextlib import suppress
from datetime import datetime
from typing import List, Optional

import aiodns  # type: ignore

from boa3.neo3 import network_logger as logger, settings, blockchain
from boa3.neo3.core import msgrouter
from boa3.neo3.network import node, payloads, convenience, message


def is_ip_address(hostname: str) -> bool:
    host = hostname.split(':')[0]
    try:
        ip = ipaddress.ip_address(host)
        return True
    except ValueError:
        return False


class NodeManager(convenience._Singleton):
    """
    This class is a convenience class that helps establish and maintain a pool of active connections to NEO nodes.

    Attention:
        This class is singleton.
    """
    #: Time interval in seconds for asking nodes for their address list.
    ADDR_QUERY_INTERVAL = 15
    #: Time interval in seconds for calling the height monitoring check.
    MONITOR_HEIGHT_INTERVAL = 30

    #: Maximum time in seconds a node may take to update its best height before being removed for being stuck.
    MAX_HEIGHT_UPDATE_DURATION = 120

    #: Time interval in seconds to check the connection pool for open spots.
    POOL_CHECK_INTERVAL = 10
    #: Maximum number of time the minimum connected client setting may be violated by the open connect checker.
    MAX_NODE_POOL_ERROR = 2
    MAX_NODE_POOL_ERROR_COUNT = 0

    #: Maximum number of times adding a block or header may fail before the node is disconnected.
    MAX_NODE_ERROR_COUNT = 5

    #: Maximum number of times a timeout threshold may be exceeded before the node is disconnected.
    #: Requires calling :meth:`~neo3.network.convenience.nodemanager.NodeManager.increase_node_timeout_count` which is
    #: done automatically by the :class:`~neo3.network.convenience.syncmanager.SyncManager` if used.
    MAX_NODE_TIMEOUT_COUNT = 15

    # init() is used instead of __init__() due to the Singleton inheritance (read its class documentation)
    def init(self):
        #: A list of nodes that we're actively using to request data from
        self.nodes = []  # type: List[node.NeoNode]
        #: A list of host:port addresses that have a task pending to to connect to, but are not fully processed
        self.queued_addresses = []

        self.tasks = []
        #: THe maximum number of clients to have connected at any time.
        self.max_clients = 5
        #: The minimum number clients to always have connected. If this is not met
        #: :attr:`~neo3.network.convenience.nodemanager.MAX_NODE_ERROR_COUNT` times then recovery logic will trigger.
        self.min_clients = 1
        self.shutting_down = False
        self.is_running = False

        msgrouter.on_node_connected += self._handle_node_connected
        msgrouter.on_node_disconnected += self._handle_node_disconnected

        self._test_client_provider = None

    def start(self) -> None:
        """
        Start the node manager services. This does 2 things

        1. Connect to the seed list addresses provided in the configuration
        2. Try to maintain a pool of connected nodes according to the `min/max clients` configuration settings and
           monitor that they don't get stuck.
        """
        def _start_services():
            self.is_running = True
            # 2. connect to addresses
            self._run_in_loop(self._fill_open_connection_spots, interval=self.POOL_CHECK_INTERVAL)

            # 3. keep trying to expand our address list with new addresses
            self._run_in_loop(self._query_addresses, interval=self.ADDR_QUERY_INTERVAL)

            # 4. monitor that connected nodes are not stuck
            self._run_in_loop(self._monitor_node_height, interval=self.MONITOR_HEIGHT_INTERVAL)

        # 1. build an initial address list
        task = asyncio.create_task(self._process_seed_list_addresses())
        task.add_done_callback(lambda fut: _start_services())

    async def shutdown(self) -> None:
        """
        Gracefully shutdown the node manager.

        Disconnects all active nodes and stops service tasks.

        Note:
            This dependents on asyncio's Task canceling logic. It waits for all tasks to be cancelled and/or stopped
            before returning.
        """
        self.shutting_down = True

        # shutdown all running tasks for this class
        # to prevent automatic filling of open spots after disconnecting nodes
        logger.debug("stopping tasks...")
        for t in self.tasks:
            t.cancel()
        await asyncio.gather(*self.tasks, return_exceptions=True)

        # finally disconnect all existing connections
        # we need to create a new list to loop over, because `disconnect` removes items from self.nodes
        to_disconnect = list(map(lambda n: n, self.nodes))
        disconnect_tasks = []
        logger.debug("disconnecting nodes...")
        for n in to_disconnect:
            disconnect_tasks.append(asyncio.create_task(n.disconnect(payloads.DisconnectReason.SHUTTING_DOWN)))
        await asyncio.gather(*disconnect_tasks, return_exceptions=True)

    """
    Start utility functions
    """

    def get_node_addresses(self) -> List[payloads.NetworkAddress]:
        """
        Return the addresses of connected nodes.
        """
        return list(map(lambda n: n.address, self.nodes))

    def get_node_by_nodeid(self, nodeid: int) -> Optional[node.NeoNode]:
        """
        Return the node instance matching the provided nodeid.
        """
        for n in self.nodes:
            if n.nodeid == nodeid:
                return n
        else:
            return None

    def get_node_with_height(self, height: int) -> Optional[node.NeoNode]:
        """
        Return a node that has a chain height of at least `height`.

        Args:
            height: the minimal block height a node must have.

        Returns:
            None if there are no nodes in the current pool that match the criteria.
            Node instance sorted by weight to avoid hitting the same node every time.
        """
        if len(self.nodes) == 0:
            return None

        weights = list(map(lambda n: n.nodeweight, self.nodes))
        # highest weight is taken first
        weights.sort(reverse=True)

        for weight in weights:
            node = self.get_node_by_nodeid(weight.id)
            if node and height <= node.best_height:
                return node
        else:
            # we could not find a node with the height we're looking for
            return None

    def get_least_failed_node(self, ri: convenience.RequestInfo) -> Optional[node.NeoNode]:
        """
        Return the node with the least fail count for the target item as specified in the RequestInfo.

        Args:
            ri: the info object indicating which header or block we filter on.

        Returns:
            None if there are no connected nodes that can provide the target information as specified in the RequestInfo
            Node instance with the lowest failure count.
        """
        # Find the node with the least failures for the item in RequestInfo

        least_failed_times = 999
        least_failed_node = None
        tried_nodes = []

        for n in self.nodes:  # type: node.NeoNode
            if n.disconnecting:
                continue

            if n.best_height < ri.height:
                continue

            failed_times = ri.failed_nodes.get(n.nodeid, 0)
            if failed_times == 0:
                # return the node we haven't tried this request on before
                return n

            tried_nodes.append(n.nodeid)
            if failed_times < least_failed_times:
                least_failed_times = failed_times
                least_failed_node = n

        return least_failed_node

    def increase_node_error_count(self, nodeid: int) -> None:
        """
        Utility function to increase a node's `error_response_count` param by 1 and disconnect the node if it exceeds
        the threshold set by :attr:`~neo3.network.convenience.nodemanager.MAX_NODE_ERROR_COUNT`.

        Args:
            nodeid (:attr:`~neo3.network.node.NeoNode.nodeid`): the specific node to update.
        """
        node = self.get_node_by_nodeid(nodeid)
        if node:
            node.nodeweight.error_response_count += 1

            if node.nodeweight.error_response_count > self.MAX_NODE_ERROR_COUNT:
                logger.debug(f"Disconnecting node {node.nodeid} Reason: max error count threshold exceeded.")
                asyncio.create_task(node.disconnect(reason=payloads.DisconnectReason.POOR_PERFORMANCE))

    def increase_node_timeout_count(self, nodeid: int) -> None:
        """
        Utility function to increase a node's `timeout_count` param by 1 and disconnect the node if it exceeds the
        threshold set by :attr:`~neo3.network.convenience.nodemanager.MAX_NODE_ERROR_COUNT`.

        Args:
            nodeid (:attr:`~neo3.network.node.NeoNode.nodeid`): the specific node to update.
        """
        node = self.get_node_by_nodeid(nodeid)
        if node:
            node.nodeweight.timeout_count += 1

            if node.nodeweight.timeout_count > self.MAX_NODE_TIMEOUT_COUNT:
                logger.debug(f"Disconnecting node {node.nodeid_human} Reason: max timeout count threshold exceeded.")
                asyncio.create_task(node.disconnect(reason=payloads.DisconnectReason.POOR_PERFORMANCE))

    """
    End utility functions
    """

    """
    Start automatic service methods.

    This section contains the logic for establishing and maintaining node connections
    """

    def _handle_node_connected(self, node: node.NeoNode) -> None:
        """ Handle node connection event."""
        self.nodes.append(node)

        with suppress(ValueError):
            self.queued_addresses.remove(node.address)

    def _handle_node_disconnected(self, node: node.NeoNode, reason: payloads.DisconnectReason) -> None:
        """ Handle node disconnection event."""
        with suppress(ValueError):
            self.nodes.remove(node)

        with suppress(ValueError):
            self.queued_addresses.remove(node.address)

    async def _process_seed_list_addresses(self) -> None:
        """
        Parses addresses from the seed list and store valid addresses
        """
        resolver = aiodns.DNSResolver(loop=asyncio.get_event_loop(), timeout=2)

        async def _process(seed):
            host, port = seed.split(':')
            if not is_ip_address(host):
                try:
                    result = await resolver.query(host, 'A')
                    # x = result[0]
                    host = result[0].host
                except aiodns.error.DNSError as e:
                    logger.debug(f"Skipping {host}, address could not be resolved: {e}.")
            node.NeoNode.addresses.append(payloads.NetworkAddress(address=f"{host}:{port}"))

        tasks = []
        for seed in settings.network.seedlist:
            tasks.append(_process(seed))
        await asyncio.gather(*tasks)

    def _connect_done_cb(self, future) -> None:
        node_instance, failure = future.result()
        # failures here are hard failures from asyncio's loop.create_connection()
        if failure:
            logger.debug(f"Failed to connect to {failure[0]} reason: {failure[1]}.")
            tmp_addr = payloads.NetworkAddress(address=failure[0])

            with suppress(ValueError):
                idx = node.NeoNode.addresses.index(tmp_addr)
                addr = node.NeoNode.addresses[idx]
                addr.set_state_dead()
                self.queued_addresses.remove(tmp_addr)

            msgrouter.on_client_connect_done(None, failure)
        else:
            msgrouter.on_client_connect_done(node_instance, None)
            node_instance.start_message_handler()

        self.tasks.remove(future)

    async def _fill_open_connection_spots(self) -> None:
        open_spots = self.max_clients - (len(self.nodes) + len(self.queued_addresses))

        if open_spots > 0:
            logger.debug(f"Found {open_spots} open pool spots, trying to add nodes...")

            # we sort the addresses such that nodes we recently disconnected from are last in the list
            # this matters in case we had to recycle addresses, meaning addresses with state POOR
            # are now labelled NEW again.
            node.NeoNode.addresses.sort(key=lambda addr: addr.last_connected)

            for _ in range(open_spots):
                # now we ask for the first address with the state NEW
                addr = node.NeoNode.get_address_new()
                if addr:
                    # an address can be queued and its state not yet changed to CONNECTED, so we must make sure we're
                    # not trying to connect to an address that is in an ongoing connection state
                    if addr not in self.queued_addresses:
                        logger.debug(f"Adding {addr} to connection queue.")
                        self.queued_addresses.append(addr)
                        if self._test_client_provider:
                            socket_mock = next(self._test_client_provider())
                            task = asyncio.create_task(node.NeoNode.connect_to(socket=socket_mock))
                        else:
                            task = asyncio.create_task(node.NeoNode.connect_to(addr.ip, addr.port))
                        self.tasks.append(task)
                        task.add_done_callback(self._connect_done_cb)
                else:
                    # oh no, we've exhausted our NEW addresses list
                    if len(self.nodes) >= self.min_clients:
                        logger.debug(f"No addresses available to fill spots. However, minimum clients still satisfied.")
                        break
                    else:
                        if self.MAX_NODE_POOL_ERROR_COUNT != self.MAX_NODE_POOL_ERROR:
                            # give our `_query_addresses` loop a chance to collect new addresses from connected nodes
                            self.MAX_NODE_POOL_ERROR_COUNT += 1
                            logger.debug(f"Increasing pool spot error count to {self.MAX_NODE_POOL_ERROR_COUNT}.")
                            break
                        else:
                            # we have no other option then to retry any address we know
                            logger.debug("Recycling old addresses.")
                            for addr in node.NeoNode.addresses:
                                if addr.is_state_poor:
                                    addr.set_state_new()
                            self.MAX_NODE_POOL_ERROR_COUNT = 0
                            break

    async def _query_addresses(self) -> None:
        """
        Ask for the address list of connected nodes on an interval.
        """
        logger.debug(f"Connected node count {len(self.nodes)}.")
        for node in self.nodes:
            logger.debug(f"Asking node {node.nodeid_human} for its address list")
            task = asyncio.create_task(node.request_address_list())
            self.tasks.append(task)
            task.add_done_callback(lambda fut: self.tasks.remove(fut))

    async def _monitor_node_height(self) -> None:
        now = datetime.utcnow().timestamp()
        for node in self.nodes:
            if now - node.best_height_last_update > self.MAX_HEIGHT_UPDATE_DURATION:
                logger.debug(f"Disconnecting node {node.nodeid} Reason: max height update threshold exceeded.")
                asyncio.create_task(node.disconnect(reason=payloads.DisconnectReason.POOR_PERFORMANCE))
            else:
                logger.debug(f"Asking node {node.nodeid_human} to send us a height update (PING)")
                # Request latest height from node
                if settings.database:
                    height = max(0, blockchain.Blockchain().height)
                else:
                    height = 0
                m = message.Message(msg_type=message.MessageType.PING, payload=payloads.PingPayload(height=height))
                task = asyncio.create_task(node.send_message(m))
                self.tasks.append(task)
                task.add_done_callback(lambda fut: self.tasks.remove(fut))

    def _run_in_loop(self, coro_func, interval) -> None:
        """
        Helper function to run a coroutine every `interval` seconds as long as the node manager is running.

        Stores a task in `tasks`

        Args:
            coro:
            interval:
        """
        async def _t(coro_func, interval):
            while not self.shutting_down:
                await coro_func()
                await asyncio.sleep(interval)

        self.tasks.append(asyncio.create_task(_t(coro_func, interval)))

    """
    End automatic service methods.
    """

    def _reset_for_test(self) -> None:
        self.nodes = []
        self.queued_addresses = []
        self.tasks = []
        self._test_client_provider = None
        self.is_running = False
